/****
Library for Raspberry Pi

:Version:   2014-03-06
:Status:    beta

:SeeAlso:   Raspberry Pi documentation, BCM2835 datasheet

:Author:    Advamation, Roland Koebler (support@advamation.de)
:Copyright: (c) Advamation (info@advamation.de)
:License:   open-source / MIT

:RCS:       $Id$
***/

//TODO:
// - blacklist / unload RPi-kernel-modules?
// - UIO / non-root-access
// - memory barriers?
// - PWM

//========================================
#include <stdio.h>      // for fopen / fclose / ...
#include <string.h>     // for strerror
#include <unistd.h>     // for close
#include <stdlib.h>     // for malloc/free
#include <sys/types.h>  // for open
#include <sys/stat.h>   // for open
#include <fcntl.h>      // for open
#include <sys/mman.h>   // for mmap / munmap
#include <time.h>       // for clock_gettime
#include <sys/file.h>   // for flock

#define G_LOG_DOMAIN "am_rpi"
#include <glib.h>

#include "advamation.h"
#include "am_rpi.h"

//========================================
// definitions

#define BCM2708_PERI_BASE   0x20000000  // RPi 1
#define BCM2709_PERI_BASE   0x3F000000  // RPi 2

#define BCM2735_TIMER   0x003000    //system timer
#define TIMER_CS        0x00/4
#define TIMER_CLO       0x04/4
#define TIMER_CHI       0x08/4
#define TIMER_C0        0x0C/4
#define TIMER_C1        0x10/4
#define TIMER_C2        0x14/4
#define TIMER_C3        0x18/4

#define BCM2735_GPIO    0x200000
#define GPIO_GPFSEL0    0x00/4  //offsets in "uint32_t"
#define GPIO_GPSET0     0x1C/4
#define GPIO_GPCLR0     0x28/4
#define GPIO_GPLEV0     0x34/4
//#define GPIO_GPEDS0     0x40/4
//#define GPIO_GPREN0     0x4C/4
//#define GPIO_GPFEN0     0x58/4
//#define GPIO_GPHEN0     0x64/4
//#define GPIO_GPLEN0     0x70/4
//#define GPIO_GPAREN0    0x7C/4
//#define GPIO_GPAFEN0    0x88/4
#define GPIO_PUD        0x94/4
#define GPIO_PUDCLK0    0x98/4

#define BCM2735_SPI     0x204000
#define SPI_CS          0x00/4
#define SPI_FIFO        0x04/4
#define SPI_CLK         0x08/4
//#define SPI_DLEN        0x0c/4
//#define SPI_LTOH        0x10/4
//#define SPI_DC          0x14/4

#define BCM2735_BSC0    0x205000    //I2C0 (on P1 at Rev. 1.0)
#define BCM2735_BSC1    0x804000    //I2C1 (on P1 at Rev. 2.0)
#define I2C_C           0x00/4
#define I2C_S           0x04/4
#define I2C_DLEN        0x08/4
#define I2C_A           0x0C/4
#define I2C_FIFO        0x10/4
#define I2C_DIV         0x14/4
#define I2C_DEL         0x18/4
#define I2C_CLKT        0x1C/4

#define PAGE_SIZE  (4*1024)
#define BLOCK_SIZE (4*1024)

//========================================
// global/static RPi

static struct {
    int init_cnt;
    uint32_t rev;

    //TODO: UIO
    int mem_fd;
    //int timer_fd:
    //int gpio_fd;
    int spi_fd;
    int i2c_fd;

    unsigned char *timer_mem;
    unsigned char *timer_map;
    uint32_t volatile *timer;

    unsigned char *gpio_mem;
    unsigned char *gpio_map;
    uint32_t volatile *gpio;

    unsigned char *spi_mem;
    unsigned char *spi_map;
    uint32_t volatile *spi;

    unsigned char *i2c_mem;
    unsigned char *i2c_map;
    uint32_t volatile *i2c;

    int      spi_lock;
    uint16_t spi_locking_timeout;
    int32_t spi_clkdiv;
    int32_t spi_cfg_anl;
    int32_t spi_cfg_orl;

    int      i2c_lock;
    uint16_t i2c_locking_timeout;
    int32_t i2c_clkdiv;
} rpi = { 0,0, -1,-1,-1, NULL,NULL,NULL, NULL,NULL,NULL, NULL,NULL,NULL, NULL,NULL,NULL,
    0,1000,-1,0xFFFFFFFF,0x00000000,
    0,1000,-1 };

/****=====================================
*    init / exit / logging
***/

/**** init library + map registers
*
* Check if it's a Raspberry Pi, initialize the library, increase the
* init-counter and map I/O memory.
* Does nothing (except increasing the init-counter), if already initialized.
*
* am_rpi_init() and am_rpi_exit() should always be called pairwise.
*
* :Returns:
*       0 on success, <0 on error
***/
AM_ERROR am_rpi_init(void)
{
    AM_ERROR ret;
    unsigned char *timer_mem_aligned;
    unsigned char *gpio_mem_aligned;
    unsigned char *spi_mem_aligned;
    unsigned char *i2c_mem_aligned;

    uint32_t bcm_peri_base;

    if(rpi.mem_fd >= 0) {
        rpi.init_cnt++;
        return 0;
    }

    //check it it's really a RPi
    ret = am_rpi_info(NULL, NULL, NULL, &(rpi.rev), NULL);
    if(ret) {
        if(ret == AM_ERROR_DEVICE_NOT_SUPPORTED)
            g_warning("System is no (supported) Raspberry Pi.");
        else
            g_warning("Cannot get hardware information. (%s)", am_error2str(ret));
        return ret;
    }
    switch(rpi.rev & 0xF0FFFF) {
        case(0xa01041): case(0xa21041):
        case(0xa02082): case(0xa22082):
            bcm_peri_base = BCM2709_PERI_BASE;
            break;
        default:
            bcm_peri_base = BCM2708_PERI_BASE;
            break;
    }

    rpi.spi_fd = -1;
    rpi.i2c_fd = -1;
    rpi.mem_fd = -1;
    rpi.timer_mem = NULL;
    rpi.timer_map = NULL;
    rpi.gpio_mem = NULL;
    rpi.gpio_map = NULL;
    rpi.spi_mem = NULL;
    rpi.spi_map = NULL;
    rpi.i2c_mem = NULL;
    rpi.i2c_map = NULL;
    rpi.spi_lock = 0;
    rpi.i2c_lock = 0;

    // lockfiles
    rpi.spi_fd = open("/var/lock/advamationrpi_spi", O_RDWR|O_CREAT, 0600);
    if(rpi.spi_fd == -1) {
        if(errno == EACCES || errno == EPERM)
            ret = AM_ERROR_PERMISSIONS;
        else
            ret = AM_ERROR_IO;
        g_warning("Cannot open lockfile /var/lock/advamationrpi_spi. (%s/%d)", strerror(errno), errno);
        goto ERROR;
    }
    rpi.i2c_fd = open("/var/lock/advamationrpi_i2c", O_RDWR|O_CREAT, 0600);
    if(rpi.i2c_fd == -1) {
        if(errno == EACCES || errno == EPERM)
            ret = AM_ERROR_PERMISSIONS;
        else
            ret = AM_ERROR_IO;
        g_warning("Cannot open lockfile /var/lock/advamationrpi_i2c. (%s/%d)", strerror(errno), errno);
        goto ERROR;
    }

    //---------------------
    // open /dev/mem
    rpi.mem_fd = open("/dev/mem", O_RDWR|O_SYNC);
    if(rpi.mem_fd < 0) {
        if(errno == EACCES || errno == EPERM)
            ret = AM_ERROR_PERMISSIONS;
        else
            ret = AM_ERROR_IO;
        g_warning("Cannot open /dev/mem. (fd %d, %s/%d)", rpi.mem_fd, strerror(errno), errno);
        goto ERROR;
    }

    //---------------------
    // mmap system timer
    rpi.timer_mem = malloc(BLOCK_SIZE + (PAGE_SIZE-1));
    if(rpi.timer_mem == NULL) {
        g_warning("Cannot allocate memory. (%s/%d)", strerror(errno), errno);
        ret = AM_ERROR_MEMORY;
        goto ERROR;
    }
    // align pointer to 4K boundary
    timer_mem_aligned = rpi.timer_mem;
    if((unsigned long)rpi.timer_mem % PAGE_SIZE)
        timer_mem_aligned += PAGE_SIZE - ((unsigned long)rpi.timer_mem % PAGE_SIZE);
    // map I/O memory
    rpi.timer_map = (unsigned char *)mmap(
      (caddr_t)timer_mem_aligned,
      BLOCK_SIZE,
      PROT_READ|PROT_WRITE,
      MAP_SHARED|MAP_FIXED,
      rpi.mem_fd,
      bcm_peri_base + BCM2735_TIMER
    );
    if((long)rpi.timer_map < 0) {
        g_warning("Cannot map memory. (%p, %s/%d)", rpi.timer_map, strerror(errno), errno);
        ret = AM_ERROR_MEMORY;
        goto ERROR;
    }
    // always use volatile pointer!
    rpi.timer = (unsigned volatile *)rpi.timer_map;

    //---------------------
    // mmap GPIO
    rpi.gpio_mem = malloc(BLOCK_SIZE + (PAGE_SIZE-1));
    if(rpi.gpio_mem == NULL) {
        g_warning("Cannot allocate memory. (%s/%d)", strerror(errno), errno);
        ret = AM_ERROR_MEMORY;
        goto ERROR;
    }
    // align pointer to 4K boundary
    gpio_mem_aligned = rpi.gpio_mem;
    if((unsigned long)rpi.gpio_mem % PAGE_SIZE)
        gpio_mem_aligned += PAGE_SIZE - ((unsigned long)rpi.gpio_mem % PAGE_SIZE);
    // map I/O memory
    rpi.gpio_map = (unsigned char *)mmap(
      (caddr_t)gpio_mem_aligned,
      BLOCK_SIZE,
      PROT_READ|PROT_WRITE,
      MAP_SHARED|MAP_FIXED,
      rpi.mem_fd,
      bcm_peri_base + BCM2735_GPIO
    );
    if((long)rpi.gpio_map < 0) {
        g_warning("Cannot map memory. (%p, %s/%d)", rpi.gpio_map, strerror(errno), errno);
        ret = AM_ERROR_MEMORY;
        goto ERROR;
    }
    // always use volatile pointer!
    rpi.gpio = (unsigned volatile *)rpi.gpio_map;

    //---------------------
    // mmap SPI
    rpi.spi_mem = malloc(BLOCK_SIZE + (PAGE_SIZE-1));
    if(rpi.spi_mem == NULL) {
        g_warning("Cannot allocate memory. (%s/%d)", strerror(errno), errno);
        ret = AM_ERROR_MEMORY;
        goto ERROR;
    }
    // align pointer to 4K boundary
    spi_mem_aligned = rpi.spi_mem;
    if((unsigned long)rpi.spi_mem % PAGE_SIZE)
        spi_mem_aligned += PAGE_SIZE - ((unsigned long)rpi.spi_mem % PAGE_SIZE);
    // map I/O memory
    rpi.spi_map = (unsigned char *)mmap(
      (caddr_t)spi_mem_aligned,
      BLOCK_SIZE,
      PROT_READ|PROT_WRITE,
      MAP_SHARED|MAP_FIXED,
      rpi.mem_fd,
      bcm_peri_base + BCM2735_SPI
    );
    if((long)rpi.spi_map < 0) {
        g_warning("Cannot map memory. (%p, %s/%d)", rpi.spi_map, strerror(errno), errno);
        ret = AM_ERROR_MEMORY;
        goto ERROR;
    }
    // always use volatile pointer!
    rpi.spi = (unsigned volatile *)rpi.spi_map;

    //---------------------
    // mmap I2C
    rpi.i2c_mem = malloc(BLOCK_SIZE + (PAGE_SIZE-1));
    if(rpi.i2c_mem == NULL) {
        g_warning("Cannot allocate memory. (%s/%d)", strerror(errno), errno);
        ret = AM_ERROR_MEMORY;
        goto ERROR;
    }
    // align pointer to 4K boundary
    i2c_mem_aligned = rpi.i2c_mem;
    if((unsigned long)rpi.i2c_mem % PAGE_SIZE)
        i2c_mem_aligned += PAGE_SIZE - ((unsigned long)rpi.i2c_mem % PAGE_SIZE);
    // map I/O memory
    if(rpi.rev == 2 || rpi.rev == 3) {  // Rev. 1.0
        rpi.i2c_map = (unsigned char *)mmap(
          (caddr_t)i2c_mem_aligned,
          BLOCK_SIZE,
          PROT_READ|PROT_WRITE,
          MAP_SHARED|MAP_FIXED,
          rpi.mem_fd,
          bcm_peri_base + BCM2735_BSC0
        );
    }
    else {                              // Rev. 2.0
        rpi.i2c_map = (unsigned char *)mmap(
          (caddr_t)i2c_mem_aligned,
          BLOCK_SIZE,
          PROT_READ|PROT_WRITE,
          MAP_SHARED|MAP_FIXED,
          rpi.mem_fd,
          bcm_peri_base + BCM2735_BSC1
        );
    }
    if((long)rpi.i2c_map < 0) {
        g_warning("Cannot map memory. (%p, %s/%d)", rpi.i2c_map, strerror(errno), errno);
        ret = AM_ERROR_MEMORY;
        goto ERROR;
    }
    // always use volatile pointer!
    rpi.i2c = (unsigned volatile *)rpi.i2c_map;

    //---------------------
    rpi.init_cnt = 1;

    return 0;
ERROR:
    am_rpi_exit();
    return ret;
}

/**** exit library + unmap registers
*
* Decrease the init-counter.
* If 0, de-initialize the library, unmap I/O memory, close logfiles and
* reset log-level.
***/
void am_rpi_exit(void) {
    if(rpi.init_cnt > 0)
        rpi.init_cnt--;

    if(rpi.init_cnt <= 0) {
        // GPIO
        rpi.gpio = NULL;
        if(rpi.gpio_map) {
            munmap(rpi.gpio_map, BLOCK_SIZE);
            rpi.gpio_map = NULL;
        }
        if(rpi.gpio_mem) {
            free(rpi.gpio_mem);
            rpi.gpio_mem = NULL;
        }
        // SPI
        if(rpi.spi_lock) {
            rpi.spi_lock = 1;
            am_rpi_spi_unlock();
        }
        rpi.spi = NULL;
        if(rpi.spi_map) {
            munmap(rpi.spi_map, BLOCK_SIZE);
            rpi.spi_map = NULL;
        }
        if(rpi.spi_mem) {
            free(rpi.spi_mem);
            rpi.spi_mem = NULL;
        }
        // I2C
        if(rpi.i2c_lock) {
            rpi.i2c_lock = 1;
            am_rpi_i2c_unlock();
        }
        rpi.i2c = NULL;
        if(rpi.i2c_map) {
            munmap(rpi.i2c_map, BLOCK_SIZE);
            rpi.i2c_map = NULL;
        }
        if(rpi.i2c_mem) {
            free(rpi.i2c_mem);
            rpi.i2c_mem = NULL;
        }

        // /dev/mem
        if(rpi.mem_fd >= 0) {
            close(rpi.mem_fd);
            rpi.mem_fd = -1;
        }
        // logging
        am_rpi_logging(G_LOG_LEVEL_WARNING, NULL, 0);
    }
}

/**** configure logging
*
* All messages up to "level" are printed and/or stored in the logfile.
* DEBUG/INFO are printed to stdout, MESSAGE/WARNING/CRITICAL/ERROR to stderr.
*
* The available levels are:
*
* - AM_LOG_LEVEL_DEBUG
* - AM_LOG_LEVEL_INFO
* - AM_LOG_LEVEL_MESSAGE
* - AM_LOG_LEVEL_WARNING (default)
* - AM_LOG_LEVEL_CRITICAL
* - AM_LOG_LEVEL_ERROR
*
* :Parameters:
*       - level: log-level for "print"
*       - logfile_level: log-level for the logfile
*       - logfile: filename of the logfile (NULL: no logfile)
* :Returns:
*       0 on success,
*       <0 on error
*       (AM_ERROR_INVALID_ARGUMENT,
*       AM_ERROR_IO if the logfile cannot be opened (see ``errno`` for details))
*
* :SeeAlso: am_logging
***/
AM_ERROR am_rpi_logging(enum am_log_level level, char const *logfile, enum am_log_level logfile_level)
{
    static FILE *f_log = NULL;
    return am_logging(G_LOG_DOMAIN, &f_log, level, logfile, logfile_level);
}

/****=====================================
*    system information
*
* :Note: These functions work even without initializing the library!
***/

static AM_ERROR _fopen_error(char const *filename) {
    AM_ERROR ret;
    if(errno == ENOMEM)  ret = AM_ERROR_MEMORY;
    if(errno == EACCES)  ret = AM_ERROR_PERMISSIONS;
    if(errno == EPERM)   ret = AM_ERROR_PERMISSIONS;
    else                 ret = AM_ERROR_IO;
    g_warning("Cannot open '%s'. (%s/%d)", filename, strerror(errno), errno);
    return ret;
}

/**** hardware information
*
* Check if this is a Raspberry Pi and get its hardware information
* from /proc/cpuinfo.
*
* :Parameters:
*       - model:        output, Raspberry Pi model
*       - rev:          output, Raspberry Pi PCB revision
*       - manufacturer: output, Raspberry Pi manufacturer
*       - rev_raw:      output, Raspberry Pi raw revision number
*       - serial:       output, Raspberry Pi serial-number
* :Returns:
*       0 on success,
*       AM_ERROR_DEVICE_NOT_SUPPORTED if this is no Raspberry Pi,
*       AM_ERROR_MEMORY, AM_ERROR_IO, AM_ERROR_PERMISSIONS on error
*
* :Uses: /proc/cpuinfo
* :Note:
*       - The following values are currently known:
*
*         ============= ==========================================
*         model         "A"
*                       "B"
*                       "A+"
*                       "B+"
*                       "Compute Module"
*                       "2B"
*                       "Zero"
*                       "3B"
*                       "UNKNOWN"
*         ------------- ------------------------------------------
*         rev           "1.0"
*                       "1.0+ECN0001/Fuses mod, D14 removed"
*                       "1.1"
*                       "1.2"
*                       "2.0 256MB"
*                       "2.0 512MB"
*                       "UNKNOWN"
*         ------------- ------------------------------------------
*         manufacturer  "Sony"
*                       "Egoman"
*                       "Quisda"
*                       "Embest"
*                       "UNKNOWN"
*         ============= ==========================================
*
*       - No memory is allocated here for model/rev/manufacturer;
*         the output pointers only point to static strings.
*       - All output-parameters may be NULL if the value is not needed.
*
* :SeeAlso: http://elinux.org/RPi_HardwareHistory
***/
AM_ERROR am_rpi_info(char const **model, char const **rev, char const **manufacturer, uint32_t *rev_raw, uint64_t *serial)
{
//TODO: overclocking-bit
//TODO: test A+, Zero
//TODO: return model-id?
    int ret, cnt;
    FILE *f;
    char *retc;
    char s[82];
    bool skip = false;
    int i;
    uint32_t r;
    long long unsigned int llu;

    //---------------------
    // /proc/cpuinfo: hardware, revision, serial
    f = fopen("/proc/cpuinfo", "r");
    if(f == NULL)  return _fopen_error("/proc/cpuinfo");

    cnt = 0;
    while(cnt < 3) {
        // read line
        retc = fgets(s, 82, f);
        if(retc == NULL || strlen(s) == 0) {
            // not all data found but EOF/error -> probably no Raspberry Pi
            //g_warning("Unexpected hardware, probably no Raspberry Pi.");
            fclose(f);
            return AM_ERROR_DEVICE_NOT_SUPPORTED;
        }
        if(strlen(s) == 81 && s[80] != '\n') {
            skip = true;
            continue;
        }
        if(skip) {
            skip = false;
            continue;
        }
        s[strlen(s)-1] = '\0';      //remove trailing newline

        // parse line
        i = 0;
        sscanf(s, "Hardware : %n", &i);
        if(i) {
            cnt++;
            if(strcmp(s+i, "BCM2708") != 0 && strcmp(s+i, "BCM2709") != 0) {
                g_warning("Unexpected hardware, probably no Raspberry Pi. ('%s')", s+i);
                fclose(f);
                return AM_ERROR_DEVICE_NOT_SUPPORTED;
            }
        }
        else if(sscanf(s, "Revision : %x", &r) == 1) {
            cnt++;
            if(rev_raw)
                *rev_raw = r;
            if(model) {
                switch(r & 0xFFFFFF) {
                    case(0x0002): case(0x0003):
                    case(0x0004): case(0x0005): case(0x0006):
                    case(0x000d): case(0x000e): case(0x000f):
                        *model = "B";
                        break;
                    case(0x0007): case(0x0008): case(0x0009):
                        *model = "A";
                        break;
                    case(0x0010):
                    case(0x0013):
                        *model = "B+";
                        break;
                    case(0x0011):
                    case(0x0014):
                        *model = "Compute Module";
                        break;
                    case(0x0012):
                        *model = "A+";
                        break;
                    case(0xa01041):
                    case(0xa21041):
                        *model = "2B";
                        break;
                    case(0x900092):
                        *model = "Zero";
                        break;
                    case(0xa02082):
                    case(0xa22082):
                        *model = "3B";
                        break;
                    default:
                        *model = "UNKNOWN";
                        break;
                }
            }
            if(rev) {
                switch(r & 0xFFFFFF) {
                    case(0x0002):
                    case(0x0010):
                    case(0x0011):
                    case(0x0012):
                    case(0x0014):
                        *rev = "1.0";
                        break;
                    case(0x0003):
                        *rev = "1.0+ECN0001/Fuses mod, D14 removed";
                        break;
                    case(0x0004): case(0x0005): case(0x0006):
                        *rev = "2.0 256MB";
                        break;
                    case(0x0007): case(0x0008): case(0x0009):
                        *rev = "2.0 256MB"; // model A
                        break;
                    case(0x000d): case(0x000e): case(0x000f):
                        *rev = "2.0 512MB";
                        break;
                    case(0x0013):
                        *rev = "1.2";
                        break;
                    case(0xa01041):
                    case(0xa21041):
                        *rev = "1.1";
                        break;
                    case(0x900092):
                        *rev = "1.2";
                        break;
                    case(0xa02082):
                    case(0xa22082):
                        *rev = "1.2";
                        break;
                    default:
                        *rev = "UNKNOWN";
                        break;
                }
            }
            if(manufacturer) {
                switch(r & 0xFFFFFF) {
                    case(0x0002): case(0x0003):
                        *manufacturer = "UNKNOWN";
                        break;
                    case(0x0004): case(0x0008): case(0x000e):
                    case(0x0010): case(0x0011): case(0x0014):
                    case(0x0012):
                    case(0xa01041):
                    case(0x900092):
                    case(0xa02082):
                        *manufacturer = "Sony";
                        break;
                    case(0x0005): case(0x0009): case(0x000f):
                        *manufacturer = "Quisda";
                        break;
                    case(0x0006): case(0x0007): case(0x000d):
                        *manufacturer = "Egoman";
                        break;
                    case(0xa21041):
                        *manufacturer = "Embest";
                        break;
                    case(0xa22082):
                    default:
                        *manufacturer = "UNKNOWN";
                        break;
                }
            }
        }
        else if(sscanf(s, "Serial : %llx", &llu) == 1) {
            cnt++;
            if(serial)
                *serial = llu;
        }
    }
    ret = fclose(f);
    if(ret)
        g_warning("fclose() '/proc/cpuinfo' failed. (%s/%d)", strerror(errno), errno);

    return 0;
}

/**** get Raspberry Pi revision-number
*
* Get Raspberry Pi revision-number.
* This reads an internal variable (and is fast after am_rpi_init()),
* and uses am_rpi_info() otherwise.
*
* :Returns:
*       0 on success, <0 on error;
*       after am_rpi_info, always 0 is returned
***/
AM_ERROR am_rpi_info_rev(uint32_t *rev_raw)
{
    if(rpi.mem_fd) {
        *rev_raw = rpi.rev;
        return 0;
    }
    else {
        return am_rpi_info(NULL, NULL, NULL, rev_raw, NULL);
    }
}

/**** CPU-clock speed information
*
* :Parameters:
*       - clk:     output, current clock in MHz
*       - clk_min: output, min. clock in MHz
*       - clk_max: output, max. clock in MHz
* :Returns:
*       0 on success,
*       AM_ERROR_DEVICE_UNEXPECTED, AM_ERROR_MEMORY, AM_ERROR_PERMISSIONS, AM_ERROR_IO on error
*
* :Note:
*       All output-parameters may be NULL if the value is not needed.
*
* :Uses: /sys/devices/system/cpu/cpu0/cpufreq/
***/
AM_ERROR am_rpi_info_clk(uint16_t *clk, uint16_t *clk_min, uint16_t *clk_max)
{
    int ret;
    FILE *f;
    long unsigned int lu;

    //---------------------
    // /sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq
    if(clk) {
        f = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq", "r");
        if(f == NULL)  return _fopen_error("/sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq");

        ret = fscanf(f, "%lu", &lu);
        fclose(f);
        if(ret == 0) {
            g_warning("Unexpected '/sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq'.");
            return AM_ERROR_DEVICE_UNEXPECTED;
        }
        *clk = lu/1000;
    }

    //---------------------
    // /sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_min_freq
    if(clk_min) {
        f = fopen("/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_min_freq", "r");
        if(f == NULL)  return _fopen_error("/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_min_freq");

        ret = fscanf(f, "%lu", &lu);
        fclose(f);
        if(ret == 0) {
            g_warning("Unexpected '/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_min_freq'.");
            return AM_ERROR_DEVICE_UNEXPECTED;
        }
        *clk_min = lu/1000;
    }

    //---------------------
    // /sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq
    if(clk_max) {
        f = fopen("/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq", "r");
        if(f == NULL)  return _fopen_error("/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq");

        ret = fscanf(f, "%lu", &lu);
        fclose(f);
        if(ret == 0) {
            g_warning("Unexpected '/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq'.");
            return AM_ERROR_DEVICE_UNEXPECTED;
        }
        *clk_max = lu/1000;
    }

    return 0;
}

/**** memory size and usage information
*
* :Parameters:
*       - mem:  output, total memory in kB
*       - used: output, used memory in kB without buffers/cache
*       - free: output, free memory in kB
*       - free_buffers_cache: output, free+buffer+cache memory in kB
* :Returns:
*       0 on success,
*       AM_ERROR_DEVICE_UNEXPECTED, AM_ERROR_MEMORY, AM_ERROR_PERMISSIONS, AM_ERROR_IO on error
*
* :Note:
*       - The returned total memory is usually less then the physical memory
*         on the board, since the memory is shared between CPU and GPU.
*       - All output-parameters may be NULL if the value is not needed.
* :Uses: /proc/meminfo
***/
AM_ERROR am_rpi_info_memory(uint32_t *mem, uint32_t *used, uint32_t *free, uint32_t *free_buffers_cache)
{
    int ret;
    FILE *f;
    long unsigned int memtotal, memfree, memavail, buffers, cached;

    f = fopen("/proc/meminfo", "r");
    if(f == NULL)  return _fopen_error("/proc/meminfo");
    ret = fscanf(f, "MemTotal: %lu kB\nMemFree: %lu kB\nBuffers: %lu kB\nCached: %lu kB\n",
            &memtotal, &memfree, &buffers, &cached);
    fclose(f);
    if(ret != 4) {
        f = fopen("/proc/meminfo", "r");
        if(f == NULL)  return _fopen_error("/proc/meminfo");
        ret = fscanf(f, "MemTotal: %lu kB\nMemFree: %lu kB\nMemAvailable: %lu kB\nBuffers: %lu kB\nCached: %lu kB\n",
                &memtotal, &memfree, &memavail, &buffers, &cached);
        fclose(f);
        if(ret != 5) {
            g_warning("Unexpected '/proc/meminfo'. (%d)", ret);
            return AM_ERROR_DEVICE_UNEXPECTED;
        }
    }

    if(mem)
        *mem = memtotal;
    if(used)
        *used = memtotal - memfree - buffers - cached;
    if(free)
        *free = memfree;
    if(free_buffers_cache)
        *free_buffers_cache = memfree + buffers + cached;

    return 0;
}

/**** temperature information
*
* :Parameters:
*       - t: output, CPU-temperature in °C
* :Returns:
*       0 on success,
*       AM_ERROR_DEVICE_UNEXPECTED, AM_ERROR_MEMORY, AM_ERROR_PERMISSIONS, AM_ERROR_IO on error
*
* :Note:
*       All output-parameters may be NULL if the value is not needed.
*
* :Uses: /sys/class/thermal/thermal_zone0/temp
***/
AM_ERROR am_rpi_info_temperature(double *t)
{
    int ret;
    FILE *f;
    long int l;

    if(t) {
        f = fopen("/sys/class/thermal/thermal_zone0/temp", "r");
        if(f == NULL)  return _fopen_error("/sys/class/thermal/thermal_zone0/temp");

        ret = fscanf(f, "%ld", &l);
        fclose(f);
        if(ret == 0) {
            g_warning("Unexpected '/sys/class/thermal/thermal_zone0/temp'.");
            return AM_ERROR_DEVICE_UNEXPECTED;
        }
        *t = ((double)l)/1000.0;
    }

    return 0;
}


/****=====================================
*    miscellaneous functions
*
* :Note: These functions work even without initializing the library!
***/

/****-------------------------------------
*    delay / wait / sleep
*
* Depending on the length of a delay, there are several ways:
*
* <= 1us:
*       Delay via busy-loop, more or less counting machine cycles.
*       Depends on CPU clock and compiler optimization, and causes
*       system load. See ``am_rpi_delay_nop*()``.
* >= 1us:
*       Delay-loop, waiting for a timer to reach the desired delay.
*       Compiler- and clock-independent, guarantees a minimum delay.
*       Causes system load. See ``am_rpi_delay_us()``.
* >= 60us:
*       Sleep-functions (e.g. nanosleep(), g_usleep()) are probably best.
*       Although they may delay the program much longer than requested,
*       they don't cause system load.
***/

/**** wait CPU-cycles
*
* Wait a specified amount of CPU cycles.
* Note that the delay is not completely predictable and depends on the
* system load and the CPU frequency.
*
* The measured values for 700 MHz, gcc -O3/-O4, idle, are roughly:
*
* === =======
* n   ns
* === =======
* 1   1.4
* 2   2.8
* 3   4.4
* 4   5.8
* 5   7.5
* 7   10.0
* 10  14.4
* 15  20.1
* 21  29.0
* 28  40.3
* 32  46.1
* 35  50.4
* 42  60.5
* 49  70.6
* 56  80.8
* 63  91.0
* 70  101.0
* 83  120.0
* === =======
*
* :Usage:
*       am_rpi_delay_nop1()
*       am_rpi_delay_nop2()
*       ...
*       am_rpi_delay_nop256()
*
***/

//see *.asm

/**** wait some µs
*
* Wait (at least) a specified amount of microseconds in a busy-loop.
*
* :Parameters:
*       - us: microseconds to wait
***/
void am_rpi_delay_us(uint16_t us)
{
    struct timespec t0, t1;
    int32_t ns;

    clock_gettime(CLOCK_MONOTONIC, &t0);
    ns = us*1000;
    while(1) {
        clock_gettime(CLOCK_MONOTONIC, &t1);
        if((t1.tv_sec-t0.tv_sec)*1000*1000*1000 + (t1.tv_nsec-t0.tv_nsec) >= ns)
            break;
    }
}


/****=====================================
*    system timer
*
* Continuously running 64-bit, 1 MHz system timer.
***/

//TODO: capture/compare registers

/**** read system timer
*
* :Returns:
*       64-bit timer value
***/
uint64_t am_rpi_timer(void)
{
    return (*(rpi.timer+TIMER_CLO)) | ((uint64_t)(*(rpi.timer+TIMER_CHI)) << 32);
}

/**** read lower-half of system timer
*
* :Returns:
*       32-bit timer value
***/
uint32_t am_rpi_timer_lo(void)
{
    return *(rpi.timer+TIMER_CLO);
}


/****=====================================
*    I/O
*
* :Note:
*   - Parameter ``gpios`` is a 32-bit-mask, containing 1 bit for every GPIO.
*     Values from ``enum am_rpi_gpio`` should be used and can be combined,
*     e.g.: ``GPIO_P1_7``, ``GPIO_P1_7|GPIO_P1_10``,
*     ``GPIO_P1_26+GPIO_P1_7+GPIO_P1_22``, ``GPIO_ALL``.
***/

/****-------------------------------------
*    I/O pins
*
* GPIO connector:
*
* ======= ============= =========== =========== =========== =========== =========== =========== ====================
* Pin     Description   ALT0        ALT1        ALT2        ALT3        ALT4        ALT5        notes
* ======= ============= =========== =========== =========== =========== =========== =========== ====================
* P1-1    (+3.3V)                                                                               (max. 50mA draw)
* P1-2    (+5V)
* P1-3    GPIO_0(Rev1)  SDA0        SA5         (res)       --          --          --          incl. 1k8 pull-up
*         GPIO_2(Rev2)  SDA1        SA3         (res)       --          --          --          incl. 1k8 pull-up
* P1-4    (+5V)
* P1-5    GPIO_1(Rev1)  SCL0        SA4         (res)       --          --          --          incl. 1k8 pull-up
*         GPIO_3(Rev2)  SCL1        SA2         (res)       --          --          --          incl. 1k8 pull-up
* P1-6    (GND)
* P1-7    GPIO_4        GPCLK0      SA1         (res)       --          --          ARM_TDI
* P1-8    GPIO_14       UART0_TXD   SD6         (res)       --          --          UART1_TXD   default: UART0_TXD
* P1-9    (GND)
* P1-10   GPIO_15       UART0_RXD   SD7         (res)       --          --          UART1_RXD   default: UART0_RXD
* P1-11   GPIO_17       (res)       SD9         (res)       UART0_RTS   SPI1_EN1_N  UART1_RTS
* P1-12   GPIO_18       PCM_CLK     SD10        (res)       sl.SDA/MOSI SPI1_CE0_N  PWM0
* P1-13   GPIO_21(Rev1) PCM_DOUT    SD13        (res)       sl.CE_N     SPI1_SCLK   GPCLK1
*         GPIO_27(Rev2) (res)       (res)       (res)       SD1_DAT3    ARM_TMS     (res)
* P1-14   (GND)
* P1-15   GPIO_22       (res)       SD14        (res)       SD1_CLK     ARM_TRST    --
* P1-16   GPIO_23       (res)       SD15        (res)       SD1_CMD     ARM_RTCK    --
* P1-17   (+3.3V)
* P1-18   GPIO_24       (res)       SD16        (res)       SD1_DAT0    ARM_TDO     --
* P1-19   GPIO_10       SPI0_MOSI   SD2         (res)       --          --          --
* P1-20   (GND)
* P1-21   GPIO_9        SPI0_MISO   SD1         (res)       --          --          --
* P1-22   GPIO_25       (res)       SD17        (res)       SD1_DAT1    ARM_TCK     --
* P1-23   GPIO_11       SPI0_SCLK   SD3         (res)       --          --          --
* P1-24   GPIO_8        SPI0_CE0_N  SD0         (res)       --          --          --
* P1-25   (GND)
* P1-26   GPIO_7        SPI0_CE1_N  SW E_N/..   (res)       --          --          --
* ------- ------------- ----------- ----------- ----------- ----------- ----------- ----------- --------------------
* P5-3    GPIO_28       SDA0        SA5         PCM_CLK     (res)       --          --
* P5-4    GPIO_29       SCL0        SA4         PCM_FS      (res)       --          --
* P5-5    GPIO_30       (res)       SA3         PCM_DIN     CTS0        --          --
* P5-6    GPIO_31       (res)       SA2         PCM_DOUT    RTS0        --          --
* ------- ------------- ----------- ----------- ----------- ----------- ----------- ----------- --------------------
* CSI-14  GPIO_2(Rev1)  SDA1        SA3         (res)       --          --          --          incl. 1k?? pull-up, default: SDA
*         GPIO_0(Rev2)  SDA0        SA5         (res)       --          --          --          incl. 1k?? pull-up, default: SDA
* CSI-13  GPIO_3(Rev1)  SCL1        SA2         (res)       --          --          --          incl. 1k?? pull-up, default: SCL
*         GPIO_1(Rev2)  SCL0        SA4         (res)       --          --          --          incl. 1k?? pull-up, default: SCL
* CSI-12  GPIO_5        GPCLK1      SA0         (res)       --          --          ARM_TDO     default: CAM_CLK
* CSI-11  GPIO_27(Rev1) (res)       (res)       (res)       SD1_DAT3    ARM_TMS     (res)       default: CAM_GPIO
*         GPIO_21(Rev2) PCM_DOUT    SD13        (res)       slave CE_N  SPI1_SCLK   GPCLK1      default: CAM_GPIO
* ======= ============= =========== =========== =========== =========== =========== =========== ====================
*
* **GPIOs are not 5V-tolerant!**
*
* :Note:
*       - Better use GPIO_P1_# instead of GPIO_#.
*       - 6 GPIOs were switched between Rev. 1.0 and Rev. 2.0
*         (GPIO_0/GPIO_2, GPIO_1/GPIO_3, GPIO_21/GPIO_27).
*         This library transparently switches the GPIOs, so that
*         all look like Rev. 2.0.
*         But better always use GPIO_P1_# instead of GPIO_#.
*       - 4 GPIOs were added on Rev. 2.0 (P5); they are ignored in Rev. 1.0.
*       - GPIO_P1_8 + GPIO_P1_10 are used by the Linux serial console by default.
*         To use them as GPIO, you have to disable the serial console and
*         reboot.
*       - For pull-up/pull-down-state on power-down, see am_rpi_gpio_pullupdown().
***/

// switch/remove GPIOs in "gpios" for Rev. 1.0
static uint32_t _gpio_remap(uint32_t gpios)
{
    if(rpi.rev == 2 || rpi.rev == 3) {
        // ignore P5
        gpios = gpios & ~(GPIO_28+GPIO_29+GPIO_30+GPIO_31);

        // GPIO_0 <-> GPIO_2
        if(     (gpios & GPIO_0) && !(gpios & GPIO_2))   gpios = (gpios & ~GPIO_0) | GPIO_2;
        else if((gpios & GPIO_2) && !(gpios & GPIO_0))   gpios = (gpios & ~GPIO_2) | GPIO_0;

        // GPIO_1 <-> GPIO_3
        if(     (gpios & GPIO_1) && !(gpios & GPIO_3))   gpios = (gpios & ~GPIO_1) | GPIO_3;
        else if((gpios & GPIO_3) && !(gpios & GPIO_1))   gpios = (gpios & ~GPIO_3) | GPIO_1;

        // GPIO_21 <-> GPIO_27
        if(     (gpios & GPIO_21) && !(gpios & GPIO_27))   gpios = (gpios & ~GPIO_21) | GPIO_27;
        else if((gpios & GPIO_27) && !(gpios & GPIO_21))   gpios = (gpios & ~GPIO_27) | GPIO_21;

    }
    return gpios;
}


/****-------------------------------------
*    I/O configuration
***/

/**** configure I/O pins
*
* :Parameters:
*       - gpios: GPIOs to configure
*       - mode:  GPIO_INPUT / GPIO_OUTPUT / GPIO_ALT0..5
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT on error
***/
AM_ERROR am_rpi_gpio_config(uint32_t gpios, enum am_rpi_gpiomode mode)
{
    g_return_val_if_fail((gpios & (~GPIO_ALL)) == 0, AM_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(mode <= 7, AM_ERROR_INVALID_ARGUMENT);
    if(gpios == 0)
        return 0;
    gpios = _gpio_remap(gpios);
    /*
    0x7E200000  GPFSEL0
    0x7E200004  GPFSEL1
    0x7E200008  GPFSEL2
    0x7E20000C  GPFSEL3
    0x7E200010  GPFSEL4
    0x7E200014  GPFSEL5
    0x7E200018  --

    reset = normal input
    unsupported_alternative = GND

    GPIO-Pins:
            SEL 0   1   2   3   4   5
        .31-30  --
        .29-27  9   19  29  ...
        .26-24  8   18  28
        .23-21  7   17  27
        .20-18  6   16  26
        .17-15  5   15  25
        .14-12  4   14  24
        .11-9   3   13  23
        .8-6    2   12  22
        .5-3    1   11  21
        .2-0    0   10  20

        000 input
        001 output
        100 alt0
        101 alt1
        110 alt2
        111 alt3
        011 alt4
        010 alt5

    => PIN pin => BYTE GPFSEL0+(pin/10), BITS (pin%10)*3+0..2
    */
    uint32_t sel_anl[4] = {0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff};
    uint32_t sel_orl[4] = {0, 0, 0, 0};
    int i;

    for(i=0; i<32 && gpios; i++) {
        if(gpios & 0x00000001) {
            sel_orl[i/10] |=  (   mode  << ((i % 10)*3));
            sel_anl[i/10] &= ~((7-mode) << ((i % 10)*3));
        }
        gpios = gpios >> 1;
    }

    if(sel_anl[0] != 0xffffffff || sel_orl[0])  *(rpi.gpio+GPIO_GPFSEL0+0) = (*(rpi.gpio+GPIO_GPFSEL0+0) & sel_anl[0]) | sel_orl[0];
    if(sel_anl[1] != 0xffffffff || sel_orl[1])  *(rpi.gpio+GPIO_GPFSEL0+1) = (*(rpi.gpio+GPIO_GPFSEL0+1) & sel_anl[1]) | sel_orl[1];
    if(sel_anl[2] != 0xffffffff || sel_orl[2])  *(rpi.gpio+GPIO_GPFSEL0+2) = (*(rpi.gpio+GPIO_GPFSEL0+2) & sel_anl[2]) | sel_orl[2];
    if(sel_anl[3] != 0xffffffff || sel_orl[3])  *(rpi.gpio+GPIO_GPFSEL0+3) = (*(rpi.gpio+GPIO_GPFSEL0+3) & sel_anl[3]) | sel_orl[3];

    return 0;
}

/**** read I/O pin configuration
*
* :Parameters:
*       - gpio: single GPIO
* :Returns:
*       enum am_rpi_gpiomode on success,
*       AM_ERROR_INVALID_ARGUMENT on error
***/
int am_rpi_gpio_config_read(enum am_rpi_gpio gpio) {
    gpio = _gpio_remap(gpio);

    int i;
    int pin = -1;
    for(i=0; i<32; i++) {
        if(gpio & 0x00000001) {
            if(gpio & 0xfffffffe) {
                g_warning("More than 1 GPIO selected. (%8.8x)", gpio << i);
                return AM_ERROR_INVALID_ARGUMENT;
            }
            pin = i;
            break;
        }
        gpio = ((unsigned)gpio) >> 1;
    }
    if(pin == -1) {
        g_warning("No GPIO selected.");
        return AM_ERROR_INVALID_ARGUMENT;
    }

    return ((*(rpi.gpio+GPIO_GPFSEL0+(pin/10))) >> ((pin%10)*3)) & 7;
}

/**** configure GPIO-pin pull-up/pull-down
*
* :Parameters:
*       - gpios: GPIOs to configure
*       - pullupdown: DISABLE / PULL_DOWN / PULL_UP
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT on error
*
* :Note:
*       On power-down, the following pull-ups/-downs are active:
*
*       - pull-up:   GPIO 0-8 / P1_3/5/7/24/26, CSI_12/13/14
*       - pull-down: GPIO 9-27, 30-31 / P1_8/10/11/12/13/15/16/18/19/21/22/23, P5_5/6, CSI_11
*       - none:      GPIO 28, 29 / P5_3/4
*
***/
AM_ERROR am_rpi_gpio_pullupdown(uint32_t gpios, enum am_rpi_pullupdown pullupdown)
{
    g_return_val_if_fail((gpios & (~GPIO_ALL)) == 0, AM_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(pullupdown <= 2, AM_ERROR_INVALID_ARGUMENT);
    if(gpios == 0)
        return 0;
    gpios = _gpio_remap(gpios);

    /*
    0x7E200094  GPPUD   GPIO Pin Pull-up/down Enable (rw)
        internal pull-up/down for ALL GPIO pins
        no read-back!
        bit 1-0
        00  disable pull-up/down
        01  pull down
        10  pull up
        11  reserved
        use with GPPUDCLK0/1
    0x7E200098  GPPUDCLK0   GPIO Pin Pull-up/down Enable Clock 0
    0x7E20009C  GPPUDCLK1   GPIO Pin Pull-up/down Enable Clock 1
        sequence:
            write GPPUD
            wait 150 cycles
            write GPPUDCLK0/1
                -> update pull-up/down for selected pins
            wait 150 cycles
            write GPPUD to remove control signal
            write GPPUDCLK0/1 to remove clock
    */
    *(rpi.gpio+GPIO_PUD) = pullupdown;
    g_usleep(1);     // min. 150 cycles (TODO: speedup?)
    *(rpi.gpio+GPIO_PUDCLK0) = gpios;
    g_usleep(1);     // min. 150 cycles (TODO: speedup?)
    *(rpi.gpio+GPIO_PUD) = 0;
    *(rpi.gpio+GPIO_PUDCLK0) = 0;

    return 0;
}


/****-------------------------------------
*    GPIO
***/

/**** read GPIO(s)
*
* :Parameters:
*       - gpios: GPIOs to read, ``0xffffffff`` for all
* :Returns:
*       read input levels, masked by ``gpios``
***/
uint32_t am_rpi_gpio_read(uint32_t gpios)
{
    /*
    0x7E200034  GPLEV0  GPIO Pin Level 0 (R)
    0x7E200038  GPLEV1  GPIO Pin Level 1 (R)
        read value
    */
    return _gpio_remap(*(rpi.gpio+GPIO_GPLEV0)) & gpios;
}

/**** write GPIO output level
*
* :Parameters:
*       - gpios: GPIOs to write
*       - level: 0/FALSE to clear, 1/TRUE to set
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT on error
***/
AM_ERROR am_rpi_gpio_write(uint32_t gpios, bool level)
{
    g_return_val_if_fail((gpios & (~GPIO_ALL)) == 0, AM_ERROR_INVALID_ARGUMENT);
    if(gpios == 0)
        return 0;
    gpios = _gpio_remap(gpios);

    /*
    0x7E20001C  GPSET0  GPIO Pin Output Set 0 (W)
        write 1 -> set for pin n
    0x7E200020  GPSET1  GPIO Pin Output Set 1 (W)
        write 1 -> set for pin n+32
    0x7E200024  --
    0x7E200028  GPCLR0  GPIO Pin Output Clear 0 (W)
        write 1 -> clear for pin n
    0x7E20002C  GPCLR1  GPIO Pin Output Clear 1 (W)
        write 1 -> clear for pin n+32
    0x7E200030  --
    */
    if(level)   *(rpi.gpio+GPIO_GPSET0) = gpios;
    else        *(rpi.gpio+GPIO_GPCLR0) = gpios;
    return 0;
}

/**** set GPIO(s)
*
* :Parameters:
*       - gpios: GPIOs to set
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT on error
***/
AM_ERROR am_rpi_gpio_set(uint32_t gpios)
{
    g_return_val_if_fail((gpios & (~GPIO_ALL)) == 0, AM_ERROR_INVALID_ARGUMENT);
    if(gpios == 0)
        return 0;
    gpios = _gpio_remap(gpios);

    *(rpi.gpio+GPIO_GPSET0) = gpios;
    return 0;
}

/**** clear GPIO(s)
*
* :Parameters:
*       - gpios: GPIOs to clear
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT on error
***/
AM_ERROR am_rpi_gpio_clr(uint32_t gpios)
{
    g_return_val_if_fail((gpios & (~GPIO_ALL)) == 0, AM_ERROR_INVALID_ARGUMENT);
    if(gpios == 0)
        return 0;
    gpios = _gpio_remap(gpios);

    *(rpi.gpio+GPIO_GPCLR0) = gpios;
    return 0;
}


/****-------------------------------------
*    SPI
*
* Broadcom BCM2835 contains 3 SPI interfaces: 1 full-featured SPI (SPI0) and
* two mini-SPIs (SPI1, SPI2). But only SPI0 is accessible on the Raspberry Pi.
*
* This library supports concurrent access to the SPI from different programs,
* which use this library, by using automatic locks:
*
* - Before every communication, the SPI is locked. If another program has
*   currently locked the SPI, the functions wait until the SPI becomes
*   unlocked or a timeout occurs.
* - After successfully locking the SPI, the SPI is reconfigured.
* - After communication, the SPI is unlocked again.
* - It's also possible to manually lock the SPI, but this should be used with
*   care, since it could prevent other programs from accessing SPI at all.
*
* If you want to use SPI from different threads in a program, use e.g.
* a mutex to make sure only 1 thread accesses the SPI at the same time.
* If the threads use different SPI-configurations, make sure to reconfigure
* the SPI after acquiring the mutex.
*
* :Note:
*       - SPI is msb-first
*       - TX- and RX-FIFO are 64 bytes.
***/

//TODO:
// - use UIO for locking
// - implement interrupts / DMA transfers?
// - implement bidir.mode / LoSSI?

/**** configure SPI-pins
*
* Use GPIO_P1_21 as MISO, GPIO_P1_19 as MOSI, GPIO_P1_23 as SCK,
* and optionally GPIO_P1_24 as CS0 / GPIO_P1_26 as CS1
*
* :Parameters:
*       - CS0: use GPIO_P1_24 as CS0 / SPI0_CE0_N?
*       - CS1: use GPIO_P1_26 as CS1 / SPI0_CE1_N?
***/
void am_rpi_spi_pins(bool cs0, bool cs1)
{
    am_rpi_gpio_config(GPIO_P1_21+GPIO_P1_19+GPIO_P1_23, GPIO_ALT0);
    if(cs0)  am_rpi_gpio_config(GPIO_P1_24, GPIO_ALT0);
    if(cs1)  am_rpi_gpio_config(GPIO_P1_26, GPIO_ALT0);
}

/**** manually lock SPI
*
* Wait until the SPI is unlocked and manually lock (and configure) the SPI.
*
* Usually, it's not necessary to use this function, since
* am_rpi_spi_start()/_stop()/_comm()/_comm_writeN()/_write_window()
* automatically lock and unlock the SPI. But the SPI may be locked manually,
* e.g. if several SPI-communications must directly follow each other, or to
* prevent other programs to access the SPI.
*
* Note that you additionally have to use a mutex, if you're accessing the
* SPI from several threads in the same program.
*
* Always use am_rpi_spi_unlock() for every successful call to this function.
*
* :Parameters:
*       - timeout_ms: timeout in ms (0..65535), 0=non-blocking, -1=use auto-locking timeout
* :Returns:
*       0 on success, AM_ERROR_BUSY after timeout
***/
AM_ERROR am_rpi_spi_lock(int32_t timeout_ms)
{
    if(rpi.spi_lock == 0) {
        uint32_t spi_cfg;
        uint32_t timeout;
        // lock
        if(flock(rpi.spi_fd, LOCK_EX|LOCK_NB) != 0) {
            if(timeout_ms < 0)
                timeout_ms = rpi.spi_locking_timeout;
            timeout = am_rpi_timer_lo() + timeout_ms*1000;
            while(1) {
                if((int32_t)(timeout - am_rpi_timer_lo()) <= 0)   // timeout
                    return AM_ERROR_BUSY;
                if(flock(rpi.spi_fd, LOCK_EX|LOCK_NB) == 0)
                    break;
                g_usleep(1);
            }
        }
        // configure
        if(rpi.spi_clkdiv >= 0)
            *(rpi.spi+SPI_CLK) = rpi.spi_clkdiv;
        spi_cfg = *(rpi.spi+SPI_CS);
        spi_cfg &= rpi.spi_cfg_anl;
        spi_cfg |= rpi.spi_cfg_orl;
        *(rpi.spi+SPI_CS) = spi_cfg;
    }
    rpi.spi_lock++;
    return 0;
}

/**** manually unlock SPI
*
* Manually unlock the SPI.
*
* This is usually only necessary after am_rpi_spi_lock(); otherwise,
* locking and unlocking is automatically done by am_rpi_spi_start()/_stop().
***/
void am_rpi_spi_unlock(void)
{
    g_assert(rpi.spi_lock);
    rpi.spi_lock--;
    if(rpi.spi_lock == 0)
        flock(rpi.spi_fd, LOCK_UN);
}

/* auto-lock + config
*/
static inline AM_ERROR _spi_lock(void)
{
    return am_rpi_spi_lock(-1);
}

/* auto-unlock
*/
static inline void _spi_unlock(void)
{
    am_rpi_spi_unlock();
}

/**** set SPI-auto-locking timeout
*
* Set timeout, after which SPI-communication-functions return AM_ERROR_BUSY
* if they cannot lock the SPI.
*
* :Parameters:
*       timeout_ms: timeout in ms (default: 1000ms)
***/
void am_rpi_spi_locking_timeout(uint16_t timeout_ms)
{
    rpi.spi_locking_timeout = timeout_ms;
}

/* fake SPI-lock
*
* If an SPI-lock is necessary over several program-runs (e.g. multiple
* am_rpi_tool runs), the lock has to be applied externally (see man 1 flock),
* and a fake-lock is necessary to prevent (auto-)locking in this library.
*
* USE WITH CARE!
*/
void am_rpi_spi_fakelock(void)
{
    if(rpi.spi_lock == 0) {
        uint32_t spi_cfg;
        // lock
        // configure
        if(rpi.spi_clkdiv >= 0)
            *(rpi.spi+SPI_CLK) = rpi.spi_clkdiv;
        spi_cfg = *(rpi.spi+SPI_CS);
        spi_cfg &= rpi.spi_cfg_anl;
        spi_cfg |= rpi.spi_cfg_orl;
        *(rpi.spi+SPI_CS) = spi_cfg;
    }
    rpi.spi_lock++;
}

/**** configure SPI
*
* :Parameters:
*       - mode: SPI mode
*       - clkcfg: SPI clock polarity and clock phase
*       - cspol0/1: SPI chip-select polarity
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT/AM_ERROR_NOT_IMPLEMENTED on error
*
* :Note:
*       - If a parameter is -1, the appr. configuration is not changed.
*       - If the SPI is currently not locked by this program, configuring the
*         SPI is delayed until the next lock.
***/
AM_ERROR am_rpi_spi_config(enum am_rpi_spi_mode mode, enum am_rpi_spi_clkcfg clkcfg, enum am_rpi_spi_cspol cspol0, enum am_rpi_spi_cspol cspol1)
{
    /*
    offset  name    descr
    0x00    CS      SPI master control+status
       .25  LEN_LONG    rw
       .24  DMA_LEN     rw
       .23  CSPOL2      rw      Chip Select 2 Polarity
       .22  CSPOL1      rw      Chip Select 1 Polarity
       .21  CSPOL0      rw      Chip Select 0 Polarity
       .20  RXF         r       RX FIFO Full
       .19  RXR         r       RX FIFO needs Reading (full)
       .18  TXD         r       TX FIFO can accept Data
       .17  RXD         r       RX FIFO contains Data
       .16  DONE        r       Data transfer Done
       .15  TE_EN               --
       .14  LMONO               --
       .13  LEN         rw      LoSSI enable
       .12  REN         rw      Read Enable [bidir. mode]
       .11  ADCS        rw      Automatically Deassert CS [DMA mode]
       .10  INTR        rw      Interrupt on RXR
       .9   INTD        rw      Interrupt on Done
       .8   DMAEN       rw      DMA Enable
       .7   TA          rw      Transfer Active
       .6   CSPOL       rw      Chip Select Polarity
       .5:4 CLEAR       rw      FIFO Clear
       .3   CPOL        rw      Clock Polarity
       .2   CPHA        rw      Clock Phase
       .1:0 CS          rw      Chip Select
    */

    switch(mode) {
        case(-1):
            break;
        case(SPI_MODE_MASTER):
            rpi.spi_cfg_anl &= ~(3 << 12);  // no LoSSI (LEN=0), no bidir (REN=0)
            rpi.spi_cfg_orl &= ~(3 << 12);
            break;
        case(SPI_MODE_BIDIR):
            g_warning("SPI mode 'BIDIR' not yet implemented.");  //TODO: implement
            return AM_ERROR_NOT_IMPLEMENTED;
        case(SPI_MODE_LoSSI):
            g_warning("SPI mode 'LoSSI' not yet implemented.");  //TODO: implement
            return AM_ERROR_NOT_IMPLEMENTED;
        default:
            g_warning("Invalid SPI mode. (%d)", mode);
            return AM_ERROR_INVALID_ARGUMENT;
    }
    switch(clkcfg) {
        case(-1):
            break;
        case(SPI_CLK_IDLELOW_SHIFTFALLING_READRISING):
            rpi.spi_cfg_anl &= ~(1<<3); rpi.spi_cfg_orl &= ~(1<<3); // idle low
            rpi.spi_cfg_anl &= ~(1<<2); rpi.spi_cfg_orl &= ~(1<<2); // shift-falling/read-rising
            break;
        case(SPI_CLK_IDLELOW_SHIFTRISING_READFALLING):
            rpi.spi_cfg_anl &= ~(1<<3); rpi.spi_cfg_orl &= ~(1<<3); // idle low
            rpi.spi_cfg_anl |=  (1<<2); rpi.spi_cfg_orl |=  (1<<2); // shift-rising/read-falling
            break;
        case(SPI_CLK_IDLEHIGH_SHIFTRISING_READFALLING):
            rpi.spi_cfg_anl |=  (1<<3); rpi.spi_cfg_orl |=  (1<<3); // idle high
            rpi.spi_cfg_anl &= ~(1<<2); rpi.spi_cfg_orl &= ~(1<<2); // shift-rising/read-falling
            break;
        case(SPI_CLK_IDLEHIGH_SHIFTFALLING_READRISING):
            rpi.spi_cfg_anl |=  (1<<3); rpi.spi_cfg_orl |=  (1<<3); // idle high
            rpi.spi_cfg_anl |=  (1<<2); rpi.spi_cfg_orl |=  (1<<2); // shift-falling/read-rising
            break;
        default:
            g_warning("Invalid SPI clock-polarity/-phase. (%d)", clkcfg);
            return AM_ERROR_INVALID_ARGUMENT;
    }
    switch(cspol0) {
        case(-1):
            break;
        case(SPI_CS_LOWACTIVE):     rpi.spi_cfg_anl &= ~(1 << 21); rpi.spi_cfg_orl &= ~(1 << 21); break;
        case(SPI_CS_HIGHACTIVE):    rpi.spi_cfg_anl |=  (1 << 21); rpi.spi_cfg_orl |=  (1 << 21); break;
        default:
            g_warning("Invalid SPI CS0-polarity. (%d)", cspol0);
            return AM_ERROR_INVALID_ARGUMENT;
    }
    switch(cspol1) {
        case(-1):
            break;
        case(SPI_CS_LOWACTIVE):     rpi.spi_cfg_anl &= ~(1 << 22); rpi.spi_cfg_orl &= ~(1 << 22); break;
        case(SPI_CS_HIGHACTIVE):    rpi.spi_cfg_anl |=  (1 << 22); rpi.spi_cfg_orl |=  (1 << 22); break;
        default:
            g_warning("Invalid SPI CS1-polarity. (%d)", cspol1);
            return AM_ERROR_INVALID_ARGUMENT;
    }

    // configure only if currently locked
    if(rpi.spi_lock) {
        uint32_t spi_cfg;
        spi_cfg = *(rpi.spi+SPI_CS);
        spi_cfg &= rpi.spi_cfg_anl;
        spi_cfg |= rpi.spi_cfg_orl;
        *(rpi.spi+SPI_CS) = spi_cfg;
    }
    return 0;
}

/**** read SPI-configuration
*
* :Parameters:
*       - mode:     output, SPI mode
*       - clkcfg:   output, SPI clock polarity and clock phase
*       - cspol0/1: output, SPI chip-select polarity
*
* :Note:
*       - This returns the current configuration, which may differ from the
*         values set by am_rpi_spi_config() if the SPI is not locked.
*       - All output-parameters may be NULL if the value is not needed.
***/
void am_rpi_spi_config_read(enum am_rpi_spi_mode *mode, enum am_rpi_spi_clkcfg *clkcfg, enum am_rpi_spi_cspol *cspol0, enum am_rpi_spi_cspol *cspol1)
{
    uint32_t spi_cfg;

    spi_cfg = *(rpi.spi+SPI_CS);

    if(mode)
        *mode = ((spi_cfg >> 12) & 0x03);
    if(clkcfg)
        *clkcfg = ((spi_cfg >> 2) & 0x03);
    if(cspol0) {
        if(spi_cfg & (1 << 21))  *cspol0 = SPI_CS_HIGHACTIVE;
        else                     *cspol0 = SPI_CS_LOWACTIVE;
    }
    if(cspol1) {
        if(spi_cfg & (1 << 22))  *cspol1 = SPI_CS_HIGHACTIVE;
        else                     *cspol1 = SPI_CS_LOWACTIVE;
    }

    return;
}

/**** set SPI clock
*
* :Parameters:
*       - clkdiv: clock divisor (SPI_CLK_* or 4/6/8/10/../65536)
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT on error
*
* :Note:
*       - SPI-speed is 250 MHz / clkdiv
*       - If the SPI is currently not locked by this program, setting the
*         SPI-clock is delayed until the next lock.
***/
AM_ERROR am_rpi_spi_clk(enum am_rpi_spi_clk clkdiv)
{
    if(clkdiv < 0)
        return 0;
    if(clkdiv < 4 || clkdiv > 65536)
        return AM_ERROR_INVALID_ARGUMENT;

    rpi.spi_clkdiv = clkdiv & 0xFFFE;

    // configure only if currently locked
    if(rpi.spi_lock)
        *(rpi.spi+SPI_CLK) = rpi.spi_clkdiv;
    return 0;
}

/**** read SPI clock
*
* :Parameters:
*       - clkdiv: output, clock divisor
*
* :Note:
*       - This returns the current clock, which may differ from the
*         value set by am_rpi_spi_clk() if the SPI is not locked.
*       - All output-parameters may be NULL if the value is not needed.
***/
void am_rpi_spi_clk_read(enum am_rpi_spi_clk *clkdiv)
{
    if(clkdiv) {
        *clkdiv = *(rpi.spi+SPI_CLK);
        if(*clkdiv == 0)
            *clkdiv = 65536;
    }
}

/**** start/init SPI
*
* Auto-lock SPI, clear FIFOs and enable SPI.
* Also reconfigure SPI if it was configured by another program.
*
* Must be called before writing any data to SPI FIFOs.
*
* :Parameters:
*       - cs: chip-select to use (0 for CS0, 1 for CS1 or -1 for none)
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT/AM_ERROR_BUSY on error
***/
AM_ERROR am_rpi_spi_start(int cs)
{
    AM_ERROR ret;

    // lock+configure SPI
    ret = _spi_lock();
    if(ret)  return ret;
    // select cs
    switch(cs) {
        case(-1):
            *(rpi.spi+SPI_CS) |= 3;  //TODO: does this work?
            break;
        case(0):
            *(rpi.spi+SPI_CS) &= ~3;
            break;
        case(1):
            *(rpi.spi+SPI_CS) = (*(rpi.spi+SPI_CS) & (~2)) | 1;
            break;
        //case(2): CS2 is not accessible on RPi
        default:
            g_warning("Invalid SPI CS. (%d)", cs);
            _spi_unlock();
            return AM_ERROR_INVALID_ARGUMENT;
    }
    // clear FIFO + set CS + enable SPI
    *(rpi.spi+SPI_CS) &= ~(1 << 7);
    *(rpi.spi+SPI_CS) |=  (1 << 7) + (3 << 4);

    return 0;
}

/**** stop/end SPI
*
* Stop and auto-unlock SPI, and clear the SPI-buffers.
* Must be called after every SPI communication.
***/
void am_rpi_spi_stop(void)
{
    *(rpi.spi+SPI_CS) &= ~(1 << 7);
    *(rpi.spi+SPI_CS) |= 0x30;
    _spi_unlock();
}

/**** write byte to SPI FIFO
*
* :Parameters:
*       - byte: byte to send
* :Returns:
*       0 on success,
*       AM_ERROR_BUFFER_FULL if the TX FIFO is full
***/
AM_ERROR am_rpi_spi_write_byte(char byte)
{
    if((*(rpi.spi+SPI_CS) & (1 << 18)) == 0)
        return AM_ERROR_BUFFER_FULL;
    *(rpi.spi+SPI_FIFO) = byte;
    return 0;
}

/**** read byte from SPI FIFO
*
* :Parameters:
*       - byte: output, received byte, may be NULL if not needed
* :Returns:
*       0 on success,
*       AM_ERROR_BUFFER_EMPTY if the RX FIFO is empty
*
* :Note: SPI is symmetric, so for each written byte, one byte must be read.
*        Otherwise the receive-buffer will run full and stop the SPI.
*        Alternatively, the SPI-RX-buffer may be cleared.
***/
AM_ERROR am_rpi_spi_read_byte(char *byte)
{
    char tmp;

    if((*(rpi.spi+SPI_CS) & (1 << 17)) == 0)
        return AM_ERROR_BUFFER_EMPTY;

    tmp = *(rpi.spi+SPI_FIFO);
    if(byte)
        *byte = tmp;

    return 0;
}

/**** write byte to SPI FIFO (blocking) + return received byte (if any)
*
* This is useful to send bytes very quickly.
*
* :Parameters:
*       - byte: byte to send
* :Returns:
*       received byte, or AM_ERROR_BUFFER_EMPTY if the RX FIFO was empty
***/
int am_rpi_spi_rw(char byte)
{
    uint32_t spi_cfg;
    int recv = AM_ERROR_BUFFER_EMPTY;

    while(1) {
        spi_cfg = *(rpi.spi+SPI_CS);
        // read 1 byte (if available)
        if(spi_cfg & (1 << 17)  &&  recv == -1)
            recv = *(rpi.spi+SPI_FIFO);
        // write byte
        if(spi_cfg & (1 << 18)) {
            *(rpi.spi+SPI_FIFO) = byte;
            break;
        }
    }
    return recv;
}

/**** wait until SPI is done
*
* Also clear the SPI-RX-buffer.
***/
void am_rpi_spi_wait_done(void)
{
    while((*(rpi.spi+SPI_CS) & (1 << 16)) == 0)
       *(rpi.spi+SPI_CS) |= 0x20;
}

/**** communicate over SPI
*
* - auto-lock + start/init SPI
* - write + read data
* - wait until done
* - stop + auto-unlock SPI
*
* :Parameters:
*       - cs: chip-select to use (0/1 or -1 for none)
*       - len: number of bytes to send/receive
*       - txdata: transmit-data
*       - rxdata: receive-memory, may be NULL
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT/AM_ERROR_BUSY on error
***/
AM_ERROR am_rpi_spi_comm(int cs, unsigned int len, char const *txdata, char *rxdata)
{
    int ret;
    unsigned int i_tx, i_rx;
    uint32_t spi_cfg;
    char tmp;

    ret = am_rpi_spi_start(cs);
    if(ret)  return ret;

    i_tx = 0;
    i_rx = 0;
    while(i_rx < len) {
        spi_cfg = *(rpi.spi+SPI_CS);
        // write byte (if still bytes to send and TX FIFO can accept data)
        if(i_tx < len  &&  spi_cfg & (1 << 18)) {
            *(rpi.spi+SPI_FIFO) = txdata[i_tx];
            i_tx++;
        }
        // read (if RX FIFO contains data)
        if(spi_cfg & (1 << 17)) {
            tmp = *(rpi.spi+SPI_FIFO);
            if(rxdata)
                rxdata[i_rx] = tmp;
            i_rx++;
        }
        //TODO: sleep if TX full and RX empty?
        //if(spi_cfg & (3 << 17) == 0) {
        //    //16 bytes @ 62.5MHz:   2.048 µs
        //    //16 bytes @ 31.25MHz:  4.096 µs
        //    //16 bytes @ 15.625MHz: 8.192 µs
        //    //...
        //    //16 bytes @ 1.95MHz:  65.5µs
        //    //16 bytes @ 0.976MHz: 131µs
        //    //...
        //    //16 bytes @ 3.81kHz:  33.6ms
        //    nanosleep(...);   // sleeps >= 90µs
        //}
    }
    am_rpi_spi_wait_done();
    am_rpi_spi_stop();
    return 0;
}

/**** communicate over SPI: write some data N times
*
* - auto-lock + start/init SPI
* - write some data (prefix)
* - write data several times
* - write some more data (suffix)
* - wait until done
* - stop + auto-unlock SPI
*
* :Parameters:
*       - cs: chip-select to use (0/1 or -1 for none)
*       - pre_len:  length of prefix
*       - pre_data: prefix-data
*       - repeat:   number of repeats
*       - len:      length of data
*       - txdata:   data
*       - suf_len:  length of suffix
*       - suf_data: suffix-data
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT/AM_ERROR_BUSY  on error
***/
AM_ERROR am_rpi_spi_comm_writeN(int cs, unsigned int pre_len, char const* pre_data, unsigned int repeat, unsigned int len, char const *txdata, unsigned int suf_len, char const *suf_data)
{
    int ret;
    unsigned int i_tx, i, N_len;
    uint32_t spi_cfg;
    volatile char tmp;
    ret = am_rpi_spi_start(cs);
    if(ret)  return ret;

    for(i=0; i<pre_len; i++)
        am_rpi_spi_rw(pre_data[i]);

    i = 0;
    i_tx = 0;
    N_len = len*repeat;
    while(i_tx < N_len) {
        spi_cfg = *(rpi.spi+SPI_CS);
        // write byte (if still bytes to send and TX FIFO can accept data)
        if(i_tx < N_len  &&  spi_cfg & (1 << 18)) {
            *(rpi.spi+SPI_FIFO) = txdata[i];
            i_tx++;
            i++;
            if(i >= len)
                i=0;
        }
        // clear RX FIFO
        *(rpi.spi+SPI_CS) |= 0x20;
        //TODO: sleep if TX full and RX empty?
    }

    for(i=0; i<suf_len; i++)
        am_rpi_spi_rw(suf_data[i]);

    am_rpi_spi_wait_done();
    am_rpi_spi_stop();
    return 0;
}

/**** communicate over SPI: write a data-window
*
* - auto-lock and start/init SPI
* - write data from a window
* - wait until done
* - stop and auto-unlock SPI
*
* :Parameters:
*       - cs: chip-select to use (0/1 or -1 for none)
*       - x0/y0: start of the data-window
*       - w: width of the data-window
*       - h: height of the data-window
*       - rowstride: length of a data-row
*       - txdata: transmit-data
*       - swap: 0=no swap, 1=swap bytes 2-byte-wise
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT/AM_ERROR_BUSY  on error
*
* :Note:
*       If swap=1, w must be even!
***/
AM_ERROR am_rpi_spi_write_window(int cs, uint16_t x0, uint16_t y0, uint16_t w, uint16_t h, uint16_t rowstride, char const *txdata, bool swap)
{
    g_return_val_if_fail(swap == 0 || swap==1, AM_ERROR_INVALID_ARGUMENT);
    if(swap==1  &&  (w&0x01))  return AM_ERROR_INVALID_ARGUMENT;

    //TODO: sleep, DMA

    int ret;
    char const *p;
    uint32_t spi_cfg;
    char tmp;
    uint_fast16_t x, y;

    ret = am_rpi_spi_start(cs);
    if(ret)  return ret;

    p = txdata + y0*rowstride + x0;
    rowstride -= w;
    for(y=0; y<h; y++) {
        x=0;
        while(x < w) {
            spi_cfg = *(rpi.spi+SPI_CS);
            // write byte (if TX FIFO can accept data)
            if(spi_cfg & (1 << 18)) {
                *(rpi.spi+SPI_FIFO) = *(p + swap);
                swap = -swap;
                x++;
                p++;
            }
            // clear RX FIFO
            *(rpi.spi+SPI_CS) |= 0x20;
        }
        p += rowstride;
    }

    am_rpi_spi_wait_done();
    am_rpi_spi_stop();
    return 0;
}


/**** SPI interface
*
* SPI interface abstraction.
*
* If a driver for an SPI-device should not be Raspberry Pi-dependent,
* better use this abstraction-layer instead of the functions above.
*
* :Usage:
*   ::
*
*       AM_ERROR ret;
*       AM_SPI_INTERFACE *my_spi;
*       my_spi = am_rpi_spi;
*       char txdata[...], rxdata[...];
*
*       ...
*
*       ret = my_spi->comm(cs, len, txdata, rxdata);
*
* :SeeAlso: documentation of am_rpi_spi_comm(), advamation.h
***/

static AM_SPI_INTERFACE const _am_rpi_spi
= {
    am_rpi_spi_comm
};
AM_SPI_INTERFACE const * const am_rpi_spi = &_am_rpi_spi;


/****-------------------------------------
*    I2C
*
* Broadcom BCM2835 contains 3 "BSC"s (Broadcom Serial Controller), which are
* compatible to fast-mode (400 Kb/s) I2C v2.1:
*
* - BSC0: on P1 on Rev. 1.0, on P5/CSI on Rev. 2.0
* - BSC1: on P1 on Rev. 2.0, on SCI on Rev. 1.0
* - BSC2: dedicated to HDMI and should not be used otherwise.
*
* Currently, only the BSC on P1 is supported.
*
* The BSC supports single-master-mode, and 7-/10-bit addressing, and has
* 16-byte-FIFOs.
*
* This library supports concurrent access to the I2C from different programs,
* which use this library:
*
* - Before every communication, the I2C is locked. If another program has
*   currently locked the I2C, the functions wait until the I2C becomes
*   unlocked or a timeout occurs.
* - After successfully locking the I2C, the I2C is reconfigured.
* - After communication, the I2C is unlocked again.
* - It's also possible to manually lock the I2C, but this should be used with
*   care, since it could prevent other programs from accessing I2C at all.
*
* If you want to use I2C from different threads in a program, use e.g.
* a mutex to make sure only 1 thread accesses the I2C at the same time.
* If the threads use different I2C-speeds, make sure to reconfigure the
* I2C-clock after acquiring the mutex.
***/

/*
I2C Registers:

    offset      name        descr
    0x00        C           control
        .31:16  --      --  (write as 0)
        .15     I2CEN   rw  I2C enable, 0=disable, 1=enable
        .14:11  --      --  (write as 0)
        .10     INTR    rw  Interrupt on RX; 1=generate int while RXR=1, 0=no int.
        .9      INTT    rw  Interrupt on TX; 1=generate int while TXW=1, 0=no int.
        .8      INTD    rw  Interrupt on Done; 1=generate int while DONE=1, 0=no int
        .7      ST      rw  start transfer; 0=no action, 1=start new transfer, one shot operation, read back as 0
        .6      --      --  (write as 0)
        .5:4    CLEAR   rw  00=no action
                            x1 / 1x =clear FIFO, one shot operation
                            (2 bits for compatibility to previous version)
                            (if CLEAR&ST: FIFO cleared before new frame)
                            read back as 0
                            clear during transfer aborts transfer
        .3:1    --      --  (write as 0)
        .0      READ    rw  read transfer
                            0=write packet transfer
                            1=read  packet transfer

    0x04    S       status
        .31:10  --      --  (write as 0)
        .9      CLKT    rw  clock stretch timeout
                            0 = no errors detected
                            1 = slave held SCL low > I2CCLKT;
                            write 1 = clear
        .8      ERR     rw  ACK error
                            0 = no errors detected
                            1 = slave has not acknowledged its address
                            write 1 = clear
        .7      RXF     r   RX-FIFO full
                            0 = FIFO not full
                            1 = FIFO is full; no further data will be received
        .6      TXE     r   TX-FIFO empty
                            0 = FIFO not empty
                            1 = FIFO empty; no further data will be transmitted
        .5      RXD     r   RX-FIFO contains data
                            0 = FIFO is empty
                            1 = FIFO contains >= 1 byte
        .4      TXD     r   TX-FIFO not full
                            0 = FIFO is full
                            1 = FIFO can accept >= 1 byte
        .3      RXR     r   RX-FIFO needs reading
                            0 = FIFO not full + read is underway
                            1 = FIFO  is full + read is underway
        .2      TXW     r   TX-FIFO needs writing
                            0 = FIFO is  full + write is underway
                            1 = FIFO not full + write is underway
        .1      DONE    rw  transfer done
                            0 = transfer not completed
                            1 = transfer comlete
                            write 1 = clear
        .0      TA      r   transfer active
                            0 = transfer not active
                            1 = transfer active

    0x08    DLEN    data length
        .31:16  --      --  (write 0)
        .15:0   DLEN    rw  data length
                            write: number of bytes to transmit/receive
                            read:  number of bytes still to transmit/receive
                            can be left over multiple packets

    0x0C    A       slave address
        .31:7   --      --  (write 0)
        .6:0    ADDR    rw  slave address

    0x10    FIFO    data fifo
        .31:8   --      --  (write 0)
        .7:0    DATA    rw  FIFO data, FIFO-size: 16 byte

    0x14    DIV     clock divisor
        .31:16  --      --  (write 0)
        .15:0   CDIV    rw  clock divisor
                            SCL = core_clock / CDIV, core_clock=250 kHz
                            rounded down to even number
                            0 = divide by 65536
                            default: 0x5dc (-> 167 kHz)
                            (The datasheet is wrong here.)

    0x18    DEL     data delay
        .31:16  FEDL    rw  falling edge delay
                            number of core clocks to wait after falling edge before outputting next bit
                            default: 0x30 (=0.32µs = 3.2% of a 100kHz-period)
        .15:0   REDL    rw  rising edge delay
                            number of core clocks to wait after falling edge before reading next bit
                            default: 0x30 (=0.32µs = 3.2% of a 100kHz-period)

    0x1C    CLKT    clock stretch timeout
        .31:16  --      --  (write 0)
        .15:0   TOUT    rw  clock stretch timeout value
                            number of SCL clock cycles to wait after the rising
                            edge of SCL before deciding that the slave is not responding
                            default: 0x40
*/
/*
TODO:
- status-checks / read status?
- implement interrupts?
- additional configuration? (CLKT, DEL, INT)
- read-back functions?
- add P5-BSC?
*/

// TODO: use UIO for locking?

/**** configure I2C-pins
*
* Use GPIO_P1_3 as SDA, GPIO_P1_5 as SCL.
***/
void am_rpi_i2c_pins(void)
{
    am_rpi_gpio_config(GPIO_P1_3+GPIO_P1_5, GPIO_ALT0);
}

/**** en-/disable I2C
*
* I2C has to be enabled before any data can be written/read.
* But the I2C-registers can also be accessed when I2C is disabled.
*
* Note that you should only disable the I2C if you are sure that no
* other application uses the I2C; otherwise, you will break other
* programs!
*
* :Parameters:
*       - enable: 0=disable I2C, 1=enable I2C
***/
void am_rpi_i2c_enable(bool enable)
{
    if(enable)
        *(rpi.i2c+I2C_C) |=   1 << 15;
    else
        *(rpi.i2c+I2C_C) &= ~(1 << 15);
}

/**** check if I2C is enabled
*
* :Returns:
*       0 if disabled, 1 if enabled
***/
bool am_rpi_i2c_enabled(void)
{
    return ((*(rpi.i2c+I2C_C)) >> 15) & 0x01;
}

/**** manually lock I2C
*
* Wait until the I2C is unlocked and manually lock (and configure) the I2C.
*
* Usually, it's not necessary to use this function, since
* am_rpi_i2c_read()/_write()/_write_read() (and its 10-bit-address variants)
* automatically lock and unlock the I2C. But the I2C may be locked manually,
* e.g. if several I2C-communications must directly follow each other, or to
* prevent other programs to access the I2C.
*
* Note that you additionally have to use a mutex, if you're accessing the
* I2C from several threads in the same program.
*
* Always use am_rpi_i2c_unlock() for every successful call to this function.
*
* :Parameters:
*       - timeout_ms: timeout in ms (0..65535), 0=non-blocking, -1=use auto-locking timeout
* :Returns:
*       0 on success, AM_ERROR_BUSY after timeout
***/
AM_ERROR am_rpi_i2c_lock(int32_t timeout_ms)
{
    if(rpi.i2c_lock == 0) {
        uint32_t timeout;
        // lock
        if(flock(rpi.i2c_fd, LOCK_EX|LOCK_NB) != 0) {
            if(timeout_ms < 0)
                timeout_ms = rpi.i2c_locking_timeout;
            timeout = am_rpi_timer_lo() + timeout_ms*1000;
            while(1) {
                if((int32_t)(timeout - am_rpi_timer_lo()) <= 0)   // timeout
                    return AM_ERROR_BUSY;
                if(flock(rpi.i2c_fd, LOCK_EX|LOCK_NB) == 0)
                    break;
                g_usleep(1);
            }
        }
        // configure
        if(rpi.i2c_clkdiv >= 0)
            *(rpi.i2c+I2C_DIV) = rpi.i2c_clkdiv;
    }
    rpi.i2c_lock++;
    return 0;
}

/**** manually unlock I2C
*
* Manually unlock the I2C.
*
* This is usually only necessary after am_rpi_i2c_lock(); otherwise, locking
* and unlocking is automatically done by am_rpi_i2c_read()/_write()/write_read().
***/
void am_rpi_i2c_unlock(void)
{
    g_assert(rpi.i2c_lock);
    rpi.i2c_lock--;
    if(rpi.i2c_lock == 0)
        flock(rpi.i2c_fd, LOCK_UN);
}

/* auto-lock + config
*/
static inline AM_ERROR _i2c_lock(void)
{
    return am_rpi_i2c_lock(-1);
}

/* auto-unlock
*/
static inline void _i2c_unlock(void)
{
    am_rpi_i2c_unlock();
}

/**** set I2C-auto-locking timeout
*
* Set timeout, after which I2C-communication-functions return AM_ERROR_BUSY
* if they cannot lock the I2C.
*
* :Parameters:
*       timeout_ms: timeout in ms (default: 1000ms)
***/
void am_rpi_i2c_locking_timeout(uint16_t timeout_ms)
{
    rpi.i2c_locking_timeout = timeout_ms;
}

/* fake I2C-lock
*
* If an I2C-lock is necessary over several program-runs (e.g. multiple
* am_rpi_tool runs), the lock has to be applied externally (see man 1 flock),
* and a fake-lock is necessary to prevent (auto-)locking in this library.
*
* USE WITH CARE!
*/
void am_rpi_i2c_fakelock(void)
{
    if(rpi.i2c_lock == 0) {
        // lock
        // configure
        if(rpi.i2c_clkdiv >= 0)
            *(rpi.i2c+I2C_DIV) = rpi.i2c_clkdiv;
    }
    rpi.i2c_lock++;
}

/**** set I2C clock
*
* :Parameters:
*       - clkdiv: clock divisor (``I2C_CLK_...kHz`` or 4/6/8/10/.../65536)
* :Returns:
*       0 on success, AM_ERROR_INVALID_ARGUMENT on error
*
* :Note:
*       - The I2C core clock is 250 MHz; default-clkdiv is 1500 (166.7kHz).
*         (The Raspberry Pi datasheet is wrong, here.)
*       - If the I2C is currently not locked by this program, setting the
*         I2C-clock is delayed until the next lock.
***/
AM_ERROR am_rpi_i2c_clk(enum am_rpi_i2c_clk clkdiv)
{
    if(clkdiv < 0)
        return 0;
    if(clkdiv < 4 || clkdiv > 65536)
        return AM_ERROR_INVALID_ARGUMENT;

    rpi.i2c_clkdiv = clkdiv & 0xFFFE;

    // configure only if currently locked
    if(rpi.i2c_lock)
        *(rpi.i2c+I2C_DIV) = rpi.i2c_clkdiv;

    return 0;
}

/**** read I2C clock
*
* :Parameters:
*       - clkdiv: output, clock divisor
*
* :Note:
*       - This returns the current configuration, which may differ from the
*         values set by am_rpi_i2c_clk() if the I2C is not locked.
*       - All output-parameters may be NULL if the value is not needed.
***/
void am_rpi_i2c_clk_read(enum am_rpi_i2c_clk *clkdiv)
{
    if(clkdiv) {
        *clkdiv = *(rpi.i2c+I2C_DIV);
        if(*clkdiv == 0)
            *clkdiv = 65536;
    }
}

/**** read byte from I2C-FIFO
*
* (low-level function without locking)
*
* :Parameters:
*       - byte: receive-memory, may be NULL
* :Returns:
*       0 on success,
*       AM_ERROR_BUFFER_EMPTY if RX FIFO is empty,
*       AM_ERROR_COMM_FAILED on timeout/ACK error
***/
AM_ERROR am_rpi_i2c_read_byte(char *byte)
{
    uint32_t s;
    char tmp;

    s = *(rpi.i2c+I2C_S);
    if( s & (3 << 8))        return AM_ERROR_COMM_FAILED;
    if((s & (1 << 5)) == 0)  return AM_ERROR_BUFFER_EMPTY;

    tmp = *(rpi.i2c+I2C_FIFO);
    if(byte)
        *byte = tmp;
    return 0;
}

/**** write byte to I2C-FIFO
*
* (low-level function without locking)
*
* :Parameters:
*       - byte: byte to send
* :Returns:
*       0 on success,
*       AM_ERROR_BUFFER_FULL if TX FIFO is full,
*       AM_ERROR_COMM_FAILED on timeout/ACK error
***/
AM_ERROR am_rpi_i2c_write_byte(char byte)
{
    uint32_t s;

    s = *(rpi.i2c+I2C_S);
    if( s & (3 << 8))        return AM_ERROR_COMM_FAILED;
    if((s & (1 << 4)) == 0)  return AM_ERROR_BUFFER_FULL;

    *(rpi.i2c+I2C_FIFO) = byte;
    return 0;
}

/**** wait until I2C is done
*
* (low-level function without locking)
*
* :Returns:
*       0 on success,
*       AM_ERROR_COMM_FAILED on timeout/ACK error
***/
AM_ERROR am_rpi_i2c_wait_done(void)
{
    uint32_t s;

    while(1) {
        s = *(rpi.i2c+I2C_S);
        if(s & (1 << 1))    return 0;
        if(s & (3 << 8))    return AM_ERROR_COMM_FAILED;
        g_usleep(1);
    }
}

/**** read data from I2C
*
* - auto-lock I2C and set I2C-speed
* - reset errors/done, set address + length, clear FIFO
* - start transfer
* - read data
* - wait until done
* - auto-unlock I2C
*
* :Parameters:
*       - address: 7-bit I2C-slave address
*       - rxlen:   number of bytes to read
*       - rxdata:  receive-memory, may be NULL
* :Returns:
*       0 on success,
*       AM_ERROR_INVALID_ARGUMENT on invalid address,
*       AM_ERROR_COMM_FAILED on timeout/ACK error,
*       AM_ERROR_BUSY on locking-timeout
***/
AM_ERROR am_rpi_i2c_read(uint8_t address, uint16_t rxlen, char *rxdata)
{
    g_return_val_if_fail(address <= 0x7f, AM_ERROR_INVALID_ARGUMENT);

    int ret;
    uint16_t i;
    char tmp;

    ret = _i2c_lock();
    if(ret)  return ret;

    *(rpi.i2c+I2C_S)    |= (3 << 8) + (1 << 1);     // reset errors + DONE
    *(rpi.i2c+I2C_A)    = address;
    *(rpi.i2c+I2C_DLEN) = rxlen;
    *(rpi.i2c+I2C_C)    |= (1 << 0);                // read
    *(rpi.i2c+I2C_C)    |= (3 << 4)  + (1 << 7);    // clear FIFO + start

    i = 0;
    while(i < rxlen) {
        ret = am_rpi_i2c_read_byte(&tmp);
        if(ret == 0) {
            if(rxdata)
                rxdata[i] = tmp;
            i++;
        }
        else if(ret == AM_ERROR_BUFFER_EMPTY) {
            // 1 byte @ 100kHz >= 90µs
            //16 byte @ 100kHz >= 1.44ms
            g_usleep(1);    //TODO: better wait?
        }
        else {
            _i2c_unlock();
            return ret; //ERROR
        }
    }

    ret = am_rpi_i2c_wait_done();
    _i2c_unlock();
    return ret;
}

/**** write data to I2C
*
* - auto-lock I2C and set I2C-speed
* - reset errors/done, set address + length + FIFO
* - start transfer
* - write more data to FIFO
* - wait until done
* - auto-unlock I2C
*
* :Parameters:
*       - address: 7-bit I2C-slave address
*       - txlen:   number of bytes to send
*       - txdata:  transmit-data
* :Returns:
*       0 on success,
*       AM_ERROR_INVALID_ARGUMENT on invalid address,
*       AM_ERROR_COMM_FAILED on timeout/ACK error,
*       AM_ERROR_BUSY on locking-timeout
***/
AM_ERROR am_rpi_i2c_write(uint8_t address, uint16_t txlen, char const *txdata)
{
    g_return_val_if_fail(address <= 0x7f, AM_ERROR_INVALID_ARGUMENT);
    //TODO: test with > 16 bytes!

    int ret;
    uint16_t i;

    ret = _i2c_lock();
    if(ret)  return ret;

    *(rpi.i2c+I2C_S)    |= (3 << 8) + (1 << 1);     // reset errors + DONE
    *(rpi.i2c+I2C_A)    = address;
    *(rpi.i2c+I2C_DLEN) = txlen;
    *(rpi.i2c+I2C_C)    &= ~(1 << 0);               // write
    *(rpi.i2c+I2C_C)    |=  (3 << 4);               // clear FIFO
    for(i=0; i < txlen && i < 16; i++)              // fill FIFO
        *(rpi.i2c+I2C_FIFO) = txdata[i];
    *(rpi.i2c+I2C_C)    |= (1 << 7);                // start

    while(i < txlen) {
        ret = am_rpi_i2c_write_byte(txdata[i]);
        if(ret == 0) {
            i++;
        }
        else if(ret == AM_ERROR_BUFFER_FULL) {
            // 1 byte @ 100kHz >= 90µs
            //16 byte @ 100kHz >= 1.44ms
            g_usleep(1);
        }
        else {
            _i2c_unlock();
            return ret; //ERROR
        }
    }

    ret = am_rpi_i2c_wait_done();
    _i2c_unlock();
    return ret;
}

/**** write data to I2C, then read data
*
* - auto-lock I2C and set I2C-speed
* - reset errors/done, set address + tx-length + FIFO
* - start transfer
* - write more data to FIFO
* - wait until TX-FIFO is empty
* - set rx-length
* - repeated start + reset done
* - read data
* - wait until done
* - auto-unlock I2C
*
* :Parameters:
*       - address: 7-bit I2C-slave address
*       - txlen:   number of bytes to send
*       - txdata:  transmit-data
*       - rxlen:   number of bytes to read
*       - rxdata:  receive-memory, may be NULL
* :Returns:
*       0 on success,
*       AM_ERROR_INVALID_ARGUMENT on invalid address,
*       AM_ERROR_COMM_FAILED on timeout/ACK error,
*       AM_ERROR_BUSY on locking-timeout
***/
AM_ERROR am_rpi_i2c_write_read(uint8_t address, uint16_t txlen, char const *txdata, uint16_t rxlen, char *rxdata)
{
    g_return_val_if_fail(address <= 0x7f, AM_ERROR_INVALID_ARGUMENT);
    //TODO: test with >= 16 byte!

    int ret;
    uint16_t i;
    uint32_t s;
    char tmp;

    ret = _i2c_lock();
    if(ret)  return ret;

    // write
    *(rpi.i2c+I2C_S)    |= (3 << 8) + (1 << 1);     // reset errors + DONE
    *(rpi.i2c+I2C_A)    = address;
    *(rpi.i2c+I2C_DLEN) = txlen;
    *(rpi.i2c+I2C_C)    &= ~(1 << 0);               // write
    *(rpi.i2c+I2C_C)    |=  (3 << 4);               // clear FIFO
    for(i=0; i < txlen && i < 16; i++)              // fill FIFO
        *(rpi.i2c+I2C_FIFO) = txdata[i];
    *(rpi.i2c+I2C_C)    |= (1 << 7);                // start

    while(i < txlen) {
        ret = am_rpi_i2c_write_byte(txdata[i]);
        if(ret == 0) {
            i++;
        }
        else if(ret == AM_ERROR_BUFFER_FULL) {
            // 1 byte @ 100kHz >= 90µs
            //16 byte @ 100kHz >= 1.44ms
            g_usleep(1);    //TODO: better wait?
        }
        else {
            _i2c_unlock();
            return ret; //ERROR
        }
    }

    // wait unitl TX-empty
    // (Note: The datasheet is wrong here, since it says to poll TA.)
    while(1) {
        s = *(rpi.i2c+I2C_S);
        if(s & (1 << 6))  break;                        //TX-FIFO empty
        if(s & (3 << 8))  { _i2c_unlock(); return AM_ERROR_COMM_FAILED; } //ERROR
        //g_usleep(1); //TODO: @100kHz: with sleep: stop+start; without: repeated start
    }

    // read with repeated start (or stop/start if not fast enough)
    *(rpi.i2c+I2C_DLEN) = rxlen;
    *(rpi.i2c+I2C_C)    |= (1 << 0);    // read
    *(rpi.i2c+I2C_C)    |= (1 << 7);    // (repeated) start
    *(rpi.i2c+I2C_S)    |= (1 << 1);    // reset DONE

    i = 0;
    while(i < rxlen) {
        ret = am_rpi_i2c_read_byte(&tmp);
        if(ret == 0) {
            if(rxdata)
                rxdata[i] = tmp;
            i++;
        }
        else if(ret == AM_ERROR_BUFFER_EMPTY) {
            // 1 byte @ 100kHz >= 90µs
            //16 byte @ 100kHz >= 1.44ms
            g_usleep(1);    //TODO: better wait?
        }
        else {
            _i2c_unlock();
            return ret; //ERROR
        }
    }

    ret = am_rpi_i2c_wait_done();
    _i2c_unlock();
    return ret;
}

/**** read data from I2C with 10-bit slave-address
*
* Like am_rpi_i2c_read(), but with 10-bit slave-address.
*
* :Parameters:
*       - address: 10-bit I2C-slave address
*       - rxlen:   number of bytes to read
*       - rxdata:  receive-memory, may be NULL
* :Returns:
*       0 on success,
*       AM_ERROR_INVALID_ARGUMENT on invalid address,
*       AM_ERROR_COMM_FAILED on timeout/ACK error,
*       AM_ERROR_BUSY on locking-timeout
***/
AM_ERROR am_rpi_i2c_read10(uint16_t address, uint16_t rxlen, char *rxdata)
{
    g_return_val_if_fail(address <= 0x3ff, AM_ERROR_INVALID_ARGUMENT);
    //TODO: test!

    /*
    10 bit addressing:
        address  = 11110XXb, XX=2 MSB of 10-bit-address
        read = write 8_LSB_of_ADDRESS, wait until transfer has started (poll TA),
               read bytes (=write DLEN, set READ + ST)
    */
    char txdata[1];
    txdata[0] = address & 0xff;
    return am_rpi_i2c_write_read(0x78 + (address>>8), 1, txdata, rxlen, rxdata);
}

/**** write data to I2C with 10-bit slave-address
*
* Like am_rpi_i2c_write(), but with 10-bit slave-address.
*
* :Parameters:
*       - address: 10-bit I2C-slave address
*       - txlen:   number of bytes to send (max. 65534)
*       - txdata:  transmit-data
* :Returns:
*       0 on success,
*       AM_ERROR_INVALID_ARGUMENT on invalid address/len,
*       AM_ERROR_COMM_FAILED on timeout/ACK error,
*       AM_ERROR_BUSY on locking-timeout
***/
AM_ERROR am_rpi_i2c_write10(uint16_t address, uint16_t txlen, char const *txdata)
{
    g_return_val_if_fail(address <= 0x3ff,  AM_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(txlen   <= 0xfffe, AM_ERROR_INVALID_ARGUMENT);
    //TODO: test!

    /*
    10 bit addressing:
        address  = 11110XXb, XX=2 MSB of 10-bit-address
        write = write 8_LSB_of_ADDRESS + data
    */
    char txdata10[65535];
    txdata10[0] = address & 0xff;
    memcpy(&(txdata10[1]), txdata, txlen);
    return am_rpi_i2c_write(0x78 + (address>>8), txlen+1, txdata10);
}

/**** write data to I2C, then read data, 10-bit slave-address
*
* Like am_rpi_i2c_write_read(), but with 10-bit slave-address.
*
* :Parameters:
*       - address: 10-bit I2C-slave address
*       - txlen:   number of bytes to send (max. 65534)
*       - txdata:  transmit-data
*       - rxlen:   number of bytes to read
*       - rxdata:  receive-memory, may be NULL
* :Returns:
*       0 on success,
*       AM_ERROR_INVALID_ARGUMENT on invalid address,
*       AM_ERROR_COMM_FAILED on timeout/ACK error,
*       AM_ERROR_BUSY on locking-timeout
***/
AM_ERROR am_rpi_i2c_write_read10(uint16_t address, uint16_t txlen, char const *txdata, uint16_t rxlen, char *rxdata)
{
    g_return_val_if_fail(address <= 0x3ff, AM_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(txlen   <= 0xfffe, AM_ERROR_INVALID_ARGUMENT);
    //TODO: test!

    char txdata10[65535];
    txdata10[0] = address & 0xff;
    memcpy(&(txdata10[1]), txdata, txlen);
    return am_rpi_i2c_write_read(0x78 + (address>>8), txlen+1, txdata10, rxlen, rxdata);
}


/**** I2C interface
*
* I2C interface abstraction.
*
* If a driver for an I2C-device should not be Raspberry Pi-dependent,
* better use this abstraction-layer instead of the functions above.
*
* :Usage:
*   ::
*
*       AM_ERROR ret;
*       AM_I2C_INTERFACE *my_i2c;
*       my_i2c = am_rpi_i2c;
*
*       ret = my_i2c->read(...);
*       ret = my_i2c->write(...);
*       ret = my_i2c->write_read(...);
*       ret = my_i2c->read10(...);
*       ret = my_i2c->write10(...);
*       ret = my_i2c->write_read10(...);
*
* :SeeAlso: documentation of the above functions, advamation.h
***/

static AM_I2C_INTERFACE const _am_rpi_i2c
= {
    am_rpi_i2c_read,
    am_rpi_i2c_write,
    am_rpi_i2c_write_read,
    am_rpi_i2c_read10,
    am_rpi_i2c_write10,
    am_rpi_i2c_write_read10
};
AM_I2C_INTERFACE const * const am_rpi_i2c = &_am_rpi_i2c;


/**** I2C-communication with Advamation RS-485-/I2C-protocol
*
* Communicate with a I2C-device with the Advamation RS-485-/I2C-protocol,
* incl. CRC/PEC.
*
* :Parameters:
*       - address: 7-bit I2C-slave address
*       - cmd:     I2C-protocol-command
*       - txlen:   number of data-bytes to send
*       - txdata:  transmit-data-bytes (0..16)
*       - rxlen:   max. number of data-bytes to read (0..16)
*       - received_len: output, number of really received bytes (may be NULL)
*       - rxdata:  receive-memory (may be NULL)
* :Returns:
*       0 on success, <0 on error
*       (AM_ERROR_INVALID_ARGUMENT on invalid address/txlen/rxlen,
*       AM_ERROR_COMM_FAILED on timeout/ACK error,
*       AM_ERROR_COMM_CRC on incorrect CRC,
*       AM_ERROR_COMM_TOOSHORT if less then rxlen bytes received),
*       AM_ERROR_BUSY on locking-timeout
*
* :SeeAlso: advamation.h, am_rpi_i2c_write_read()
***/
AM_ERROR am_rpi_i2c_prot(uint8_t address, enum am_cmd cmd, unsigned int txlen, char const *txdata, unsigned int rxlen, unsigned int *received_len, char *rxdata)
{
    if(address >= 0x80 || txlen > 16 || rxlen > 16)
        return AM_ERROR_INVALID_ARGUMENT;

    AM_ERROR ret;
    char s[19], r[18];
    char crc;
    int i;

    //TODO: optimize?

    // build send-data
    s[0] = txlen + 1;
    s[1] = cmd;
    crc = am_crc_smbus(0,   (address << 1));
    crc = am_crc_smbus(crc, s[0]);
    crc = am_crc_smbus(crc, s[1]);
    for(i=0; (unsigned)i<txlen; i++) {
        s[i+2] = txdata[i];
        crc = am_crc_smbus(crc, txdata[i]);     // update CRC
    }
    s[txlen+2] = crc;

    // communicate
    ret = am_rpi_i2c_write_read(address, txlen+3, s, rxlen+2, r);
    if(ret)
        return ret;

    // parse received data + CRC
    if(r[0] > 16)
        // TODO: if 0xFF: other error?
        return AM_ERROR_COMM_TOOLARGE;
    crc = am_crc_smbus(0,   (address << 1) + 1);
    crc = am_crc_smbus(crc, r[0]);
    if(received_len)
        *received_len = r[0];
    for(i=0; i<r[0]; i++) {
        if(rxdata  &&  rxlen > (unsigned)i)
            rxdata[i] = r[i+1];
        crc = am_crc_smbus(crc, r[i+1]);
    }

    if(crc != r[r[0]+1])
        return AM_ERROR_COMM_CRC;
    if((unsigned)r[0] != rxlen)
        return AM_ERROR_COMM_TOOSHORT;

    return 0;
}


/****=====================================
*    UI-functions
***/

//---------------------
// mappings
static const struct am_int2str gpio_strings[] = {
    { GPIO_P1_3,  "GPIO_P1_3"  },
    { GPIO_P1_5,  "GPIO_P1_5"  },
    { GPIO_P1_7,  "GPIO_P1_7"  },
    { GPIO_P1_8,  "GPIO_P1_8"  },
    { GPIO_P1_10, "GPIO_P1_10" },
    { GPIO_P1_11, "GPIO_P1_11" },
    { GPIO_P1_12, "GPIO_P1_12" },
    { GPIO_P1_13, "GPIO_P1_13" },
    { GPIO_P1_15, "GPIO_P1_15" },
    { GPIO_P1_16, "GPIO_P1_16" },
    { GPIO_P1_18, "GPIO_P1_18" },
    { GPIO_P1_19, "GPIO_P1_19" },
    { GPIO_P1_21, "GPIO_P1_21" },
    { GPIO_P1_22, "GPIO_P1_22" },
    { GPIO_P1_23, "GPIO_P1_23" },
    { GPIO_P1_24, "GPIO_P1_24" },
    { GPIO_P1_26, "GPIO_P1_26" },

    { GPIO_P5_3,  "GPIO_P5_3"  },   // Rev. 2.0
    { GPIO_P5_4,  "GPIO_P5_4"  },   // Rev. 2.0
    { GPIO_P5_5,  "GPIO_P5_5"  },   // Rev. 2.0
    { GPIO_P5_6,  "GPIO_P5_6"  },   // Rev. 2.0

    { GPIO_CSI_11, "GPIO_CSI_11" }, // GPIOs on CSI connector
    { GPIO_CSI_12, "GPIO_CSI_12" },
    { GPIO_CSI_13, "GPIO_CSI_13" },
    { GPIO_CSI_14, "GPIO_CSI_14" },

    { GPIO_2,   "GPIO_2"  },
    { GPIO_3,   "GPIO_3"  },
    { GPIO_4,   "GPIO_4"  },
    { GPIO_7,   "GPIO_7"  },
    { GPIO_8,   "GPIO_8"  },
    { GPIO_9,   "GPIO_9"  },
    { GPIO_10,  "GPIO_10" },
    { GPIO_11,  "GPIO_11" },
    { GPIO_14,  "GPIO_14" },
    { GPIO_15,  "GPIO_15" },
    { GPIO_17,  "GPIO_17" },
    { GPIO_18,  "GPIO_18" },
    { GPIO_22,  "GPIO_22" },
    { GPIO_23,  "GPIO_23" },
    { GPIO_24,  "GPIO_24" },
    { GPIO_25,  "GPIO_25" },
    { GPIO_27,  "GPIO_27" },
    { GPIO_28,  "GPIO_28" },    // Rev. 2.0
    { GPIO_29,  "GPIO_29" },    // Rev. 2.0
    { GPIO_30,  "GPIO_30" },    // Rev. 2.0
    { GPIO_31,  "GPIO_31" },    // Rev. 2.0
    { GPIO_0,   "GPIO_0" },     // CSI
    { GPIO_1,   "GPIO_1" },     // CSI
    { GPIO_5,   "GPIO_5" },     // CSI
    { GPIO_21,  "GPIO_21" },    // CSI

    { GPIO_P1, "GPIO_P1" },
    { GPIO_P5, "GPIO_P5" },
    { GPIO_CSI, "GPIO_CSI" },
    { GPIO_ALL, "GPIO_ALL" },
    { 0, NULL }
};

/**** map enum GPIO_* to a string
*
* :Returns:
*       string or "unknown"
***/
const char* am_rpi_gpio2str(enum am_rpi_gpio gpio)
{
    const char *str;
    str = am_int2str_lookup(gpio_strings, gpio, TRUE);
    if(str)  return str;
    else     return "unknown";
}

/**** map string to enum GPIO_*
*
* :Returns:
*       found enum or -1
***/
enum am_rpi_gpio am_rpi_str2gpio(char const *s)
{
    return am_int2str_rlookup(gpio_strings, s, -1);
}

//========================================

