/****
Commandline-tool for Raspberry Pi

:Version:   2013-11-19
:Status:    beta

:Author:    Advamation, Roland Koebler (support@advamation.de)
:Copyright: (c) Advamation (info@advamation.de)
:License:   open-source / MIT

:RCS:       $Id$
***/

#define VERSION "2015-03-16"
#define AUTHOR "Advamation (support@advamation.de)"

//========================================
//GLib
#include <glib.h>

#include <stdio.h>
#include <string.h>     // for strcmp
#include <inttypes.h>

#include "advamation.h"
#include "am_rpi.h"

//========================================
static gchar *mode=NULL;
static gint verbose=G_LOG_LEVEL_INFO;
static gboolean fakelock=FALSE;

static GOptionEntry entries[] = {
    {"mode",    'm', 0, G_OPTION_ARG_STRING, &mode, "\n\
        INFO                    - print system information\n\
        TIMER                   - print time of 64-bit, 1 MHz system timer\n\
\n\
        GPIO_CONFIG      GPIOs MODE       - configure GPIO pins\n\
        GPIO_CONFIG_READ                  - read GPIO pin configuration\n\
        GPIO_PULLUPDOWN  GPIOs PULLUPDOWN - configure GPIO pull-up/down\n\
        GPIO_READ        [GPIOs]          - read GPIOs\n\
        GPIO_WRITE       GPIOs LEVEL      - write GPIOs\n\
        GPIO_SET         GPIOs            - set   GPIOs\n\
        GPIO_CLR         GPIOs            - clear GPIOs\n\
\n\
            MODE:       INPUT, OUTPUT, ALT0..5\n\
            GPIOs:      GPIO_0..25 / GPIO_P1_3..26 / GPIO_P5_3..6 /\n\
                        GPIO_P1 / GPIO_P5 / GPIO_ALL / 0x... / 0b...\n\
            PULLUPDOWN: 0=disable, 1=pull down, 2=pull up\n\
            LEVEL:      0=low, 1=high\n\
\n\
        SPI_CONFIG      MODE CLKCFG [CSPOL0] [CSPOL1] - configure SPI\n\
        SPI_CONFIG_READ                               - read SPI-configuration\n\
        SPI_CLK         CLKDIV                        - set SPI-clock\n\
        SPI_CLK_READ                                  - read SPI-clock\n\
        SPI_COMM        CS DATA...                    - comm. over SPI\n\
\n\
            MODE:       -1=don't change, 0=master, 1=bidir. 2=LoSSI\n\
            CLKCFG:     -1=don't change\n\
                        0=idle low,  shift falling/read rising\n\
                        1=idle low,  shift rising/read falling\n\
                        2=idle high, shift rising/read falling (*)\n\
                        3=idle high, shift falling/read rising\n\
            CSPOL0/1:   -1=don't change, 0=low active, 1=high active\n\
            CLKDIV:     clock-divider 2/4/6/8/10..65536, SPI-clk = 250MHz/CLKDIV\n\
            CS:         -1=none, 0=CS0, 1=CS1\n\
            DATA:       1..16 data bytes\n\
\n\
        I2C_CONFIG      ENABLE          - en/disable I2C + configure GPIO pins\n\
        I2C_CLK         CLK             - set I2C-clock\n\
        I2C_CLK_READ                    - read I2C-clock\n\
        I2C_READ        ADDRESS N_READ           - read\n\
        I2C_WRITE       ADDRESS DATA...          - write\n\
        I2C_COMM        ADDRESS N_READ DATA...   - write, then read\n\
        I2C_READ10      ADDRESS10 N_READ         - read,       10-bit-addr.\n\
        I2C_WRITE10     ADDRESS10 DATA....       - write,      10-bit-addr.\n\
        I2C_COMM10      ADDRESS10 N_READ DATA... - write+read, 10-bit-addr.\n\
\n\
            ENABLE:     0=disable, 1=enable\n\
            CLK:        clock-divider 2/4/8...65536, I2C-clk = 250MHz/CLK, or\n\
                        5/10/20/50/100/200/300/400kHz\n\
            ADDRESS:    7-bit-address of I2C-slave\n\
            ADDRESS10:  10-bit-address of I2C-slave\n\
            N_READ:     number of bytes to read\n\
            DATA:       1..16 data bytes to write\n\
\n\
        All parameters are case-insensitive.\n\
        All parameters can be decimal, hexadecimal (0x...) or binary (0b...)\n\
        numbers.\n\
        If you want to use \"-1\" as parameter, prepend it with \"--\" to make\n\
        sure that it's not interpreted as an option, e.g.\n\
        ... -m SPI_CONFIG -- 0 -1 -1 1\n\
", "MODE" },
    {"nolock",  'l', 0, G_OPTION_ARG_NONE, &fakelock, "don't use locking (use with care!)", NULL},
    {"verbose", 'v', 0, G_OPTION_ARG_INT, &verbose, "verbosity level (default: 64)\n\
      (DEBUG: 128, INFO: 64, MESSAGE: 32, WARNING: 16, CRITICAL: 8, ERROR: 4/0)", "LEVEL"},
    {NULL,0,0,0,NULL,NULL,NULL}
};

//========================================
int main(int argc, char **argv)
{
    int ret=0;
    GError *gerror=NULL;

    //---------------------
    //commandline-parsing
    GOptionContext *context;

    context = g_option_context_new("- Advamation Raspberry Pi commandline-interface");
    g_option_context_add_main_entries(context, entries, NULL);
    ret = g_option_context_parse(context, &argc,&argv, &gerror);
    if(verbose >= G_LOG_LEVEL_INFO)
        g_print("Advamation RaspberryPi tool, V%s, (c) %s\n\n", VERSION, AUTHOR);
    if(!ret) {
        g_printerr("ERROR: Cannot parse command line: %s\n", gerror->message);
        g_option_context_free(context);
        g_error_free(gerror);
        goto ERROR;
    }
    g_option_context_free(context);

    if(!mode)
        goto USAGE;

    //---------------------
    // verbosity
    if(verbose < 4)
        verbose = 4;
    if(verbose >= 16 && verbose <= 64) {
        am_logging(NULL,NULL, 8, NULL,0);
        am_rpi_logging(8, NULL,0);
    }
    else {
        am_logging(NULL,NULL, verbose, NULL,0);
        am_rpi_logging(verbose, NULL, 0);
    }

    //---------------------
    // parse parameters/data
    //
    // 0x...
    // 0b...
    // [0-9]+
    // -[0-9]+
    // GPIO_*
    #define N_TOKENS 30
    int      i, n_param;
    uint32_t u32;
    int32_t  i32;
    char     c;
    int64_t  param_i[N_TOKENS];
    char*    param_s[N_TOKENS];

    n_param = 0;
    for(i=1; (int)i<argc; i++) {
        // skip "--"
        if(strcmp(argv[i], "--") == 0)
            continue;

        if(n_param >= N_TOKENS) {
            g_printerr("ERROR: Too many data-bytes (currently max. %d tokens).\n", N_TOKENS);
            goto ERROR;
        }
        param_s[n_param] = argv[i];

        //0x...
        if(strlen(argv[i]) > 2  &&  argv[i][0] == '0'  && argv[i][1] == 'x') {
            ret = sscanf(argv[i], "0x%x%c", &u32, &c);
            if(ret != 1)  {g_printerr("ERROR: Invalid hex data '%s'.\n", argv[i]);  goto ERROR;}
            param_i[n_param] = u32;
        }
        //0b...
        else if(strlen(argv[i]) > 2  &&  argv[i][0] == '0'  && argv[i][1] == 'b') {
            unsigned int j;
            param_i[n_param] = 0;
            for(j=2; j<strlen(argv[i]); j++) {
                param_i[n_param] = param_i[n_param] << 1;
                if(     argv[i][j] == '0')  {}
                else if(argv[i][j] == '1')  {param_i[n_param] |= 1;}
                else {g_printerr("ERROR: Invalid binary data '%s'.\n", argv[i]);  goto ERROR;}
            }
        }
        //negative number
        else if(strlen(argv[i]) > 1  &&  argv[i][0] == '-'  &&  sscanf(argv[i], "%d%c", &i32, &c) == 1) {
            param_i[n_param] = i32;
        }
        //unsigned number
        else if(strlen(argv[i]) >= 1  &&  sscanf(argv[i], "%u%c", &u32, &c) == 1) {
            param_i[n_param] = u32;
        }
        //string
        else {
            // GPIO_*
            ret = am_rpi_str2gpio(argv[i]);
            if(ret >= 0)
                param_i[n_param] = ret;
            else {
                param_i[n_param] = -2;
            }
        }
        n_param++;
    }
    //print parsed data
    if(verbose >= G_LOG_LEVEL_DEBUG) {
        for(i=0; i<n_param; i++) {
            g_print("%d: ", i);
            g_print("0x%08"PRIx64" %10"PRId64" ", param_i[i]&0xffffffff, param_i[i]);
            am_print_uint_bin(param_i[i], 32);
            g_print(" '%s'", param_s[i]);
            g_print("\n");
        }
    }

    //---------------------
    // init library
    if(g_ascii_strcasecmp("INFO", mode) != 0) {
        ret = am_rpi_init();
        if(ret) {
            g_printerr("ERROR: Cannot init library. (%d, %s)\n", ret, am_error2str(ret));
            goto ERROR;
        }
        if(fakelock) {
            am_rpi_spi_fakelock();
            am_rpi_i2c_fakelock();
        }
    }

    //---------------------
    // modes

    //----------------------------------------
    // info
    if(g_ascii_strcasecmp("INFO", mode) == 0) {
        if(n_param != 0)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        char const *model, *rev, *manuf;
        uint16_t clk, clk_min, clk_max;
        uint32_t rev_raw;
        uint64_t serial;
        uint32_t mem, memused, memfree, memfree2;
        double t;

        ret = am_rpi_info(&model, &rev, &manuf, &rev_raw, &serial);
        if(ret != 0)  goto CLEANUP;
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("\
Model:        %s\n\
Revision:     %s (0x%x)\n\
Manufacturer: %s\n\
Serial:       0x%"PRIx64" / %"PRId64"\n\
", model, rev, rev_raw, manuf, serial, serial);
        }
        else {
            g_print("%s, %d, %s, 0x%"PRIx64"\n", model, rev_raw, manuf, serial);
        }

        ret = am_rpi_info_clk(&clk, &clk_min, &clk_max);
        if(ret != 0)  goto CLEANUP;
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("\n\
CPU-Clock:    %3.3d MHz (%3.3d..%3.3d MHz)\n\
", clk, clk_min, clk_max);
        }
        else {
            g_print("%3.3d MHz, ", clk);
        }

        ret = am_rpi_info_temperature(&t);
        if(ret != 0)  goto CLEANUP;
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("\
CPU-Temp.:    %4.1lf ", t);
            printf("°"); //TODO: fix this
            g_print("C\n");
        }
        else {
            g_print("%4.1lf", t);
            printf("°"); //TODO: fix this
            g_print("C\n");
        }

        ret = am_rpi_info_memory(&mem, &memused, &memfree, &memfree2);
        if(ret != 0)  goto CLEANUP;
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("\n\
Memory:          %7.3lf MB\n\
Memory used-b-c: %7.3lf MB\n\
Memory free:     %7.3lf MB\n\
Memory free+b+c: %7.3lf MB\n\
", (double)mem/1000, (double)memused/1000, (double)memfree/1000, (double)memfree2/1000);
        }
        else {
            g_print("%.3lf MB / %.3lf MB / %.3lf MB / %.3lf MB\n", (double)mem/1000, (double)memused/1000, (double)memfree/1000, (double)memfree2/1000);
        }
    }
    //----------------------------------------
    // timer
    else if(g_ascii_strcasecmp("TIMER", mode) == 0) {
        if(n_param != 0)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        uint64_t t;
        t = am_rpi_timer();
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("TIMER: %"PRIu64"us (%"PRIu64" days, %02"PRIu64":%02"PRIu64":%02"PRIu64".%08"PRIu64")\n", t, t/1000000/60/60/24, (t/1000000/60/60)%24, (t/1000000/60)%60, (t/1000000)%60, t%10000000);
        }
        else {
            g_print("%"PRIu64"\n", t);
        }
    }
    //----------------------------------------
    // GPIO
    else if(g_ascii_strcasecmp("GPIO_CONFIG", mode) == 0) {
        if(n_param != 2)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        if(     g_ascii_strcasecmp(param_s[1], "INPUT") == 0)  param_i[1] = GPIO_INPUT;
        else if(g_ascii_strcasecmp(param_s[1], "OUTPUT")== 0)  param_i[1] = GPIO_OUTPUT;
        else if(g_ascii_strcasecmp(param_s[1], "ALT0")  == 0)  param_i[1] = GPIO_ALT0;
        else if(g_ascii_strcasecmp(param_s[1], "ALT1")  == 0)  param_i[1] = GPIO_ALT1;
        else if(g_ascii_strcasecmp(param_s[1], "ALT2")  == 0)  param_i[1] = GPIO_ALT2;
        else if(g_ascii_strcasecmp(param_s[1], "ALT3")  == 0)  param_i[1] = GPIO_ALT3;
        else if(g_ascii_strcasecmp(param_s[1], "ALT4")  == 0)  param_i[1] = GPIO_ALT4;
        else if(g_ascii_strcasecmp(param_s[1], "ALT5")  == 0)  param_i[1] = GPIO_ALT5;
        if(param_i[1] < 0  ||  param_i[1] > 7) {
            g_printerr("ERROR: Invalid MODE.\n");
            goto ERROR;
        }

        ret = am_rpi_gpio_config(param_i[0], param_i[1]);
    }
    else if(g_ascii_strcasecmp("GPIO_CONFIG_READ", mode) == 0) {
        if(n_param != 0)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        int i;
        uint32_t rev;
        am_rpi_info_rev(&rev);
        for(i=0; i<32; i++) {
            if(i==6 || i==12 || i==13 || i==16 || i==19 || i==20 || i==26)
                continue;
            if((rev < 4) && i > 27)
                break;
#if __AM_RPI_USE_CSI_GPIO
#else
            if(i==0 || i==1 || i==5 || i==21)
                continue;
#endif
            ret = am_rpi_gpio_config_read(1 << i);
            if(verbose >= G_LOG_LEVEL_MESSAGE) {
                g_print("GPIO_%-2d / %-11s: %x", i, am_rpi_gpio2str(1<<i), ret);
                switch(ret) {
                    case(GPIO_INPUT):    g_print(" (INPUT)"); break;
                    case(GPIO_OUTPUT):   g_print(" (OUTPUT)"); break;
                    case(GPIO_ALT0):     g_print(" (ALT0)"); break;
                    case(GPIO_ALT1):     g_print(" (ALT1)"); break;
                    case(GPIO_ALT2):     g_print(" (ALT2)"); break;
                    case(GPIO_ALT3):     g_print(" (ALT3)"); break;
                    case(GPIO_ALT4):     g_print(" (ALT4)"); break;
                    case(GPIO_ALT5):     g_print(" (ALT5)"); break;
                }
                g_print("\n");
            }
            else {
                g_print("%d ", ret);
            }
        }
        if(verbose < G_LOG_LEVEL_MESSAGE)
            g_print("\n");
        ret = 0;
    }
    else if(g_ascii_strcasecmp("GPIO_PULLUPDOWN", mode) == 0) {
        if(n_param != 2)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        ret = am_rpi_gpio_pullupdown(param_i[0], param_i[1]);
    }
    else if(g_ascii_strcasecmp("GPIO_READ", mode) == 0) {
        if(n_param == 0) {
            n_param = 1;
            param_i[0] = 0xffffffff;
        }
        if(n_param != 1)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        uint32_t values;
        values = am_rpi_gpio_read(param_i[0]);

        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            int i;
            g_print("GPIOs: 0x%8.8x\n", values);
            for(i=0; i<30; i++) {
                if(i==6 || i==12 || i==13 || i==16 || i==19 || i==20 || i==26)
                    continue;
#if __AM_RPI_USE_CSI_GPIO
#else
                if(i==0 || i==1 || i==5 || i==21)
                    continue;
#endif
                if(param_i[0] & (1 << i)) {
                    g_print("GPIO_%-2d / %-11s: %d\n", i, am_rpi_gpio2str(1<<i), (values & (1<<i)) ? 1 : 0);
                }
            }
        }
        else {
            g_print("0x%8.8x\n", values);
        }
        ret = 0;
    }
    else if(g_ascii_strcasecmp("GPIO_WRITE", mode) == 0) {
        if(n_param != 2)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}
        ret = am_rpi_gpio_write(param_i[0], param_i[1]);
    }
    else if(g_ascii_strcasecmp("GPIO_SET", mode) == 0) {
        if(n_param != 1)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}
        ret = am_rpi_gpio_set(param_i[0]);
    }
    else if(g_ascii_strcasecmp("GPIO_CLR", mode) == 0) {
        if(n_param != 1)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}
        ret = am_rpi_gpio_clr(param_i[0]);
    }
    //----------------------------------------
    // SPI
    else if(g_ascii_strcasecmp("SPI_CONFIG", mode) == 0) {
        if(n_param != 2 && n_param != 3 && n_param != 4)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        ret = am_rpi_spi_lock(-1);
        if(ret)  goto CLEANUP;
        am_rpi_spi_pins(n_param>=3 ? TRUE : FALSE, n_param==4 ? TRUE : FALSE);
        ret = am_rpi_spi_config(param_i[0], param_i[1],
                n_param>=3 ? param_i[2] : SPI_CS_DONTCHANGE,
                n_param==4 ? param_i[3] : SPI_CS_DONTCHANGE);
        am_rpi_spi_unlock();
    }
    else if(g_ascii_strcasecmp("SPI_CONFIG_READ", mode) == 0) {
        if(n_param != 0)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        enum am_rpi_spi_mode mode;
        enum am_rpi_spi_clkcfg clkcfg;
        enum am_rpi_spi_cspol cspol0, cspol1;
        am_rpi_spi_config_read(&mode, &clkcfg, &cspol0, &cspol1);
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("SPI mode:           %s\n", mode == SPI_MODE_MASTER ? "MASTER" : (mode == SPI_MODE_BIDIR ? "BIDIR" : (mode == SPI_MODE_LoSSI ? "LoSSI" : "unknown")));
            g_print("SPI clock-polarity: %s\n", clkcfg <= 1 ? "idle low" : "idle HIGH");
            g_print("SPI clock-phase:    %s\n", (clkcfg == 0 || clkcfg == 3) ? "shift at falling-edge, read at rising-edge" : "shift at rising-edge, read at falling-edge");
            g_print("SPI CS0-polarity:   %s\n", cspol0 ? "HIGH-active" : "low-active");
            g_print("SPI CS1-polarity:   %s\n", cspol1 ? "HIGH-active" : "low-active");
        }
        else {
            g_print("%s, ", mode == SPI_MODE_MASTER ? "MASTER" : (mode == SPI_MODE_BIDIR ? "BIDIR" : (mode == SPI_MODE_LoSSI ? "LoSSI" : "unknown")));
            g_print("%s/",   clkcfg <= 1 ? "idle low" : "idle HIGH");
            g_print("%s, ", (clkcfg == 0 || clkcfg == 3) ? "shift at falling-edge/read at rising-edge" : "shift at rising-edge/read at falling-edge");
            g_print("CS0 %s, ", cspol0 ? "HIGH-active" : "low-active");
            g_print("CS1 %s\n", cspol1 ? "HIGH-active" : "low-active");
        }
    }
    else if(g_ascii_strcasecmp("SPI_CLK", mode) == 0) {
        if(n_param != 1)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        ret = am_rpi_spi_lock(-1);
        if(ret)  goto CLEANUP;
        ret = am_rpi_spi_clk(param_i[0]);
        am_rpi_spi_unlock();
    }
    else if(g_ascii_strcasecmp("SPI_CLK_READ", mode) == 0) {
        if(n_param != 0)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        enum am_rpi_spi_clk clkdiv;
        am_rpi_spi_clk_read(&clkdiv);
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            if(clkdiv <= 250)
                g_print("SPI clock: %.2f MHz (clkdiv %d)\n", 250.0/clkdiv, clkdiv);
            else
                g_print("SPI clock: %.2f kHz (clkdiv %d)\n", 250000.0/clkdiv, clkdiv);
        }
        else {
            g_print("%d\n", clkdiv);
        }
    }
    else if(g_ascii_strcasecmp("SPI_COMM", mode) == 0) {
        if(n_param < 2)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}
        if(n_param > 17)  {g_printerr("ERROR: Too large/not implemented number of tx-bytes.\n"); goto ERROR;}

        char txdata[16];
        char rxdata[16];
        int i;

        for(i=1; i<n_param; i++)
            txdata[i-1] = (char)(param_i[i]);
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("->");
            for(i=0; i<n_param-1; i++)  g_print(" 0x%2.2x", txdata[i]);
            g_print("\n");
        }
        ret = am_rpi_spi_comm(param_i[0], n_param-1, txdata, rxdata);
        if(verbose >= G_LOG_LEVEL_MESSAGE)
            g_print("<- ");
        for(i=0; i<n_param-1; i++)
            g_print("0x%2.2x ", rxdata[i]);
        g_print("\n");
    }
    //----------------------------------------
    // I2C
    else if(g_ascii_strcasecmp("I2C_CONFIG", mode) == 0) {
        if(n_param != 1)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        if(param_i[0]) {
            am_rpi_i2c_pins();
            am_rpi_i2c_enable(true);
        }
        else {
            am_rpi_i2c_enable(false);
        }
        ret = 0;
    }
    else if(g_ascii_strcasecmp("I2C_CLK", mode) == 0) {
        if(n_param != 1)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        enum am_rpi_i2c_clk clkdiv;
        if(param_i[0] < 0) {
            if(     g_ascii_strcasecmp(param_s[0], "5kHz")   == 0)  clkdiv = I2C_CLK_5kHz;
            else if(g_ascii_strcasecmp(param_s[0], "10kHz")  == 0)  clkdiv = I2C_CLK_10kHz;
            else if(g_ascii_strcasecmp(param_s[0], "20kHz")  == 0)  clkdiv = I2C_CLK_20kHz;
            else if(g_ascii_strcasecmp(param_s[0], "50kHz")  == 0)  clkdiv = I2C_CLK_50kHz;
            else if(g_ascii_strcasecmp(param_s[0], "100kHz") == 0)  clkdiv = I2C_CLK_100kHz;
            else if(g_ascii_strcasecmp(param_s[0], "200kHz") == 0)  clkdiv = I2C_CLK_200kHz;
            else if(g_ascii_strcasecmp(param_s[0], "300kHz") == 0)  clkdiv = I2C_CLK_300kHz;
            else if(g_ascii_strcasecmp(param_s[0], "400kHz") == 0)  clkdiv = I2C_CLK_400kHz;
            else {g_printerr("ERROR: Invalid parameter 'CLK' (%s).\n", param_s[0]); goto ERROR;}
        }
        else {
            clkdiv = param_i[0];
        }
        ret = am_rpi_i2c_lock(-1);
        if(ret)  goto CLEANUP;
        ret = am_rpi_i2c_clk(clkdiv);
        am_rpi_i2c_unlock();
    }
    else if(g_ascii_strcasecmp("I2C_CLK_READ", mode) == 0) {
        if(n_param != 0)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}

        enum am_rpi_i2c_clk clkdiv;
        am_rpi_i2c_clk_read(&clkdiv);
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            if(clkdiv <= 250)
                g_print("I2C clock: %.2f MHz (clkdiv %d)\n", 250.0/clkdiv, clkdiv);
            else
                g_print("I2C clock: %.2f kHz (clkdiv %d)\n", 250000.0/clkdiv, clkdiv);
        }
        else {
            g_print("%d\n", clkdiv);
        }
    }
    else if(g_ascii_strcasecmp("I2C_WRITE", mode) == 0) {
        if(n_param < 2)   {g_printerr("ERROR: Wrong number of parameters.\n");  goto ERROR;}
        if(n_param > 17)  {g_printerr("ERROR: Too large/not implemented number of tx-bytes.\n"); goto ERROR;}

        char txdata[16];
        for(i=1; i<n_param; i++)
            txdata[i-1] = param_i[i];
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("->");
            for(i=0; i<n_param-1; i++)  g_print(" 0x%2.2x", txdata[i]);
            g_print("\n");
        }
        ret = am_rpi_i2c_write(param_i[0], n_param-1, txdata);
    }
    else if(g_ascii_strcasecmp("I2C_READ", mode) == 0) {
        if(n_param != 2)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}
        if(param_i[1] < 0  ||  param_i[1] > 16) {
            g_printerr("ERROR: Invalid/not implemented number of rx-bytes. (%"PRId64")\n", param_i[1]);
            goto ERROR;
        }

        char rxdata[16];
        ret = am_rpi_i2c_read(param_i[0], param_i[1], rxdata);
        if(verbose >= G_LOG_LEVEL_MESSAGE)
            g_print("<- ");
        for(i=0; i<param_i[1]; i++)
            g_print("0x%2.2x ", rxdata[i]);
        g_print("\n");
    }
    else if(g_ascii_strcasecmp("I2C_COMM", mode) == 0) {
        if(n_param < 3)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}
        if(n_param > 18)  {g_printerr("ERROR: Too large/not implemented number of tx-bytes.\n"); goto ERROR;}
        if(param_i[1] < 1  ||  param_i[1] > 16) {
            g_printerr("ERROR: Invalid/not implemented number of rx-bytes. (%"PRId64")\n", param_i[1]);
            goto ERROR;
        }

        char rxdata[16];
        char txdata[16];
        for(i=2; i<n_param; i++)
            txdata[i-2] = param_i[i];
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("->");
            for(i=0; i<n_param-2; i++)  g_print(" 0x%2.2x", txdata[i]);
            g_print("\n");
        }
        ret = am_rpi_i2c_write_read(param_i[0], n_param-2, txdata, param_i[1], rxdata);
        if(verbose >= G_LOG_LEVEL_MESSAGE)
            g_print("<- ");
        for(i=0; i<param_i[1]; i++)
            g_print("0x%2.2x ", rxdata[i]);
        g_print("\n");
    }
    else if(g_ascii_strcasecmp("I2C_WRITE10", mode) == 0) {
        if(n_param < 2)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}
        if(n_param > 17)  {g_printerr("ERROR: Too large/not implemented number of tx-bytes.\n"); goto ERROR;}

        char txdata[16];
        for(i=1; i<n_param; i++)
            txdata[i-1] = param_i[i];
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("->");
            for(i=0; i<n_param-1; i++)  g_print(" 0x%2.2x", txdata[i]);
            g_print("\n");
        }
        ret = am_rpi_i2c_write10(param_i[0], n_param-1, txdata);
    }
    else if(g_ascii_strcasecmp("I2C_READ10", mode) == 0) {
        if(n_param != 2)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}
        if(param_i[1] < 0  ||  param_i[1] > 16) {
            g_printerr("ERROR: Invalid/not implemented number of rx-bytes. (%"PRId64")\n", param_i[1]);
            goto ERROR;
        }

        char rxdata[16];
        ret = am_rpi_i2c_read10(param_i[0], param_i[1], rxdata);
        if(verbose >= G_LOG_LEVEL_MESSAGE)
            g_print("<- ");
        for(i=0; i<param_i[1]; i++)
            g_print("0x%2.2x ", rxdata[i]);
        g_print("\n");
    }
    else if(g_ascii_strcasecmp("I2C_COMM10", mode) == 0) {
        if(n_param < 3)   {g_printerr("ERROR: Wrong number of parameters.\n"); goto ERROR;}
        if(n_param > 18)  {g_printerr("ERROR: Too large/not implemented number of tx-bytes.\n"); goto ERROR;}
        if(param_i[1] < 1  ||  param_i[1] > 16) {
            g_printerr("ERROR: Invalid/not implemented number of rx-bytes. (%"PRId64")\n", param_i[1]);
            goto ERROR;
        }

        char rxdata[16];
        char txdata[16];
        for(i=2; i<n_param; i++)
            txdata[i-2] = param_i[i];
        if(verbose >= G_LOG_LEVEL_MESSAGE) {
            g_print("->");
            for(i=0; i<n_param-2; i++)  g_print(" 0x%2.2x", txdata[i]);
            g_print("\n");
        }
        ret = am_rpi_i2c_write_read10(param_i[0], n_param-2, txdata, param_i[1], rxdata);
        if(verbose >= G_LOG_LEVEL_MESSAGE)
            g_print("<- ");
        for(i=0; i<param_i[1]; i++)
            g_print("0x%2.2x ", rxdata[i]);
        g_print("\n");
    }
    //----------------------------------------
    // invalid mode
    else {
        g_printerr("ERROR: Invalid mode.\n");
        goto ERROR;
    }

    goto CLEANUP;
USAGE:
    g_printerr("use --help for usage information\n");
    ret = -1;
ERROR:
    if(ret == 0)
        ret = -1;
CLEANUP:
    am_rpi_exit();

    if(ret == -1)
        ;
    else if(ret < 0)
        g_printerr("ERROR: %s (%d)\n", am_error2str(ret), ret);
    else if(verbose >= G_LOG_LEVEL_MESSAGE)
        g_print("OK.\n");
    return ret;
}

//========================================

